/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef NBC_DEF_H
#define NBC_DEF_H

#ifdef VBNC
#undef VBNC
#endif

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <math.h>
#ifdef USE_ICONV
#  include <iconv.h>
#endif
#include <errno.h>
#include <stdarg.h>
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
#  include <sys/time.h>
#  include <sys/resource.h>
#  include <unistd.h>
#endif
#include "xgetopt.h"
#ifdef _OPENMP
#  include <omp.h>
#endif

#if __STD_VERSION__ >= 199901L
#  define CURR_FUNCTION __func__
#elif defined(_MSC_VER) || defined(__GNUC__)
#  define CURR_FUNCTION __FUNCTION__
#else
#  define CURR_FUNCTION (__FILE__ ": " __LINE__)
#endif

#define TOOL_NAME "NBCT"
#ifndef VERSION
#  define VERSION "(unknown)"
#endif
#define DEFAULT_CLASS_SMOOTH (0)
#define DEFAULT_WORD_SMOOTH  (0)
#define DEFAULT_MAX_ITERATE  (3000)
#define DEFAULT_RAND_SEED    ((unsigned int)time(NULL))
#define HASH_SIZE            (997)
#define HASH_RADIX           (97)
#define HASH_EXPAND          (0.75)
#define MAX_WORD_LEN         (256)
#define MAX_FREQ_LEN         (30)
#define MAX_INT_LEN          (30)
#define MAX_FLOAT_LEN        (60)
#define MAX_LINE_LEN         (131072)
#define MAX_OUT_LEN          (1048576)
#define MAX_FSX_LEN          (16)
#define MAX_FEXT_LEN         (16)
#define ENUM_TYPE_LEN        (64)
#define CHARCODE_LEN         (100)
#ifndef MAX_MAX_ITERATE
#  define MAX_MAX_ITERATE    (10000)
#endif
#define MAX_FILENAME         (512)
#define MAX_FILENAME_AUX     (2 * MAX_FILENAME + MAX_FSX_LEN + MAX_FEXT_LEN + 8)
#define INIT_TMP_CLASS       (50)
#define INIT_TMP_WORD        (1000)
#define DEFAULT_SMOOTH_S     (1.0)
#define DEFAULT_EPSILON      (1.0e-3)
#define DEFAULT_ITEMP_INIT   (0.1)
#define DEFAULT_ITEMP_RATE   (1.2)
#ifndef ACCEPT_SMALL_NUMBER
#  define ACCEPT_SMALL_NUMBER (1.0e-6)
#endif
#define NOISE_RATIO         (1.0e-1)
#define NOISE_RATIO_GAUSS   (1.0e-1)
#define IGNORE_JSPLIT       (1.0e-6)
#define SMALL_FLOAT         (1.0e-6)
#define VERY_SMALL_FLOAT    (1.0e-15)
#define TINY_FLOAT          (1.0e-30)
#define MIN_SMEM_IMPROVE_RATIO (10.0)
#define DEFAULT_MIN_PARAM_C (1.0e-3)
#define DEFAULT_MIN_NEW_PARAM_C (1.0e-3)
#define SIGMA_LIMIT         (1.0e-12)
#define ALPHA_W_NOISE       (1.0e-6)
#define Q_REGULAR_MIN       (0.05)
#define Q_REGULAR_MAX       (0.95)
#define CHARCODE_INTERNAL "UTF-8"
#define FSX_FREQ    "freq"
#define FSX_PARAM   "param"
#define FSX_HPARAM  "hparam"
#define FSX_DIST    "dissim"
#define FSX_MEMP    "memp"
#define FSX_LOG     "log"
#define FSX_MSG     "msg"
#define FSX_CLUSTER "cluster"
#define FSX_RANK    "rank"
#define FSX_EVAL    "eval"
#define SMOOTH_AUTO       "auto"
#define INIT_STR_RANDOM   "random"
#define INIT_STR_NOISY_U  "noisy_u"
#define INIT_STR_VFREQ    "v_freq"
#define SMOOTH_VAR_STR_SAMPLE "sample"
#define SMOOTH_VAR_STR_CONST  "const"
#define IC_STR_BIC        "BIC"
#define IC_STR_AUTOCLASS  "Cheeseman-Stutz score"
#define FEXT_TSV          "txt"
#define FEXT_CSV          "csv"
#define SEP_GAUSS         '#'
#define RANGE_GAUSS_MV    '-'
#define RANGE_GAUSS_UV    '~'
#define RANGE_IGNORE      '-'
/* the values below are should be negative */
#define WORD_NOT_ASSIGNED (-1)
#define WORD_MISSING      (-2)
#define DEFAULT_PREC_OUT  12
#define DEFAULT_PREC_MSG  6
#define MAX_PREC_OUT      36  /* should be enough even for long double */
#define MAX_PREC_MSG      36
#define INT2STR_BUFSIZE   12  /* = log10(INT_MAX) + 2 */
#define FLOAT2STR_BUFSIZE 60
#define NUM_FIELD_CLASS   2
#define NUM_FIELD_CAT     4
#define NUM_FIELD_MEAN    4
#define NUM_FIELD_COVAR   5
#define NUM_FIELD_TAU     2
#define NUM_FIELD_ALPHA   3
#define LU_SCALE_LEN      256
#define INIT_NUM_M_BLOCK  256
#define INIT_NUM_V_BLOCK  256

#ifdef NO_LDOUBLE
typedef double FLOAT_T;
#else
typedef long double FLOAT_T;
#endif

#ifdef ICONV_CONST
typedef const char *ICONV_BUF_TYPE;
#else
typedef char *ICONV_BUF_TYPE;
#endif

#ifdef USE_CRLF
#  define NEWLINE(fp) do { fputs("\r\n",(fp)); } while (0)
#else
#  define NEWLINE(fp) do { fputs("\n",(fp));   } while (0)
#endif

#ifndef M_PI
# ifdef NO_LDOUBLE
#  define M_PI 3.14159265358979323846
# else
#  define M_PI 3.14159265358979323846L
# endif
#endif

#ifdef _MSC_VER
#  define POW(x,y)  pow((double)(x),(double)(y))
#  define EXP(x)    exp((double)(x))
#  define SQRT(x)   sqrt((double)(x))
#  define LOG(x)    log((double)(x))
#  define FABS(x)   fabs((double)(x))
#elif defined(NO_LDOUBLE)
#  define POW(x,y)  pow((x),(y))
#  define EXP(x)    exp(x)
#  define SQRT(x)   sqrt(x)
#  define LOG(x)    log(x)
#  define FABS(x)   fabs(x)
#elif defined(HAVE_EXPL)
#  define POW(x,y)  powl((x),(y))
#  define EXP(x)    expl(x)
#  define SQRT(x)   sqrtl(x)
#  define LOG(x)    logl(x)
#  define FABS(x)   fabsl(x)
#else
#  define POW(x,y)  pow((double)(x),(double)(y))
#  define EXP(x)    exp((double)(x))
#  define SQRT(x)   sqrt((double)(x))
#  define LOG(x)    log((double)(x))
#  define FABS(x)   fabs((double)(x))
#endif

#ifdef _MSC_VER
#  define STRTOD(s,p)  strtod((s),(p))
#elif defined(NO_LDOUBLE)
#  define STRTOD(s,p)  strtod((s),(p))
#elif defined(HAVE_STRTOLD)
#  define STRTOD(s,p)  strtold((s),(p))
#else
#  define STRTOD(s,p)  ((FLOAT_T)strtod((s),(p)))
#endif

/* C99 is assumed */
#ifdef _MSC_VER
#  define ISNAN(x)  _isnan((double)(x))
#  define FINITE(x) _finite((double)(x))
#else
#  define ISNAN(x)  isnan(x)
#  define FINITE(x) isfinite(x)
#endif
#define ISINF(x) (!FINITE(x))

/* Use the Cholesky decomposition or not */
#ifdef NO_CHOLESKY
#  define INVERT(x,y)    invert(x,y)
#  define DETERMINANT(x) determinant(x)
#else
#  define INVERT(x,y)    invert_c(x,y)
#  define DETERMINANT(x) determinant_c(x)
#endif

#define LOG2(x)   (LOG(x) / LOG(2))
#define LOG10(x)  (LOG(x) / LOG(10))

/* Choose(n,2), the number of possible pairs from n objects */
#define CHOOSE2(n) ((n >= 2) ? (n) * ((n) - 1) / 2 : 0)

/* macros for error handling */
#define DIE_LINE_NFIELD(fname,lc,buf,msg)  \
  die_line((fname),(lc),__LINE__,(buf),(msg),-1)
#define DIE_LINE_ILL_INT(fname,lc,buf)     \
  die_line((fname),(lc),__LINE__,(buf),"illegal format for an integer",-1)
#define DIE_LINE_ILL_FLOAT(fname,lc,buf)   \
  die_line((fname),(lc),__LINE__,(buf),    \
	   "illegal format for an floating-point number",-1)
#define DIE_NUMERIC(msg) die_numeric(CURR_FUNCTION,(msg))
#define DIE_DIM          die_dimension(CURR_FUNCTION)

/* Sometimes used for VB routines (this makes no problem if it is
 * sufficient to get a string representation of the option) */
#define OPTSTR(i)  get_option_str((i),nbc_opts,0)
#define LOPTSTR(i) get_option_str((i),nbc_opts,1)

#define xmalloc(s)     xmalloc0((s),     __FILE__,__LINE__)
#define xrealloc(p,s)  xrealloc0((p),(s),__FILE__,__LINE__)
#define alloc_str(l)   alloc_str0((l),   __FILE__,__LINE__)
#define calloc_str(l)  calloc_str0((l),  __FILE__,__LINE__)
#define dup_str(s)     dup_str0((s),     __FILE__,__LINE__)
#define xfree(s)           do { free(s); s = NULL; } while (0)
#define release_vector(v)  do { release_vector0(v); v = NULL; } while (0)
#define release_matrix(m)  do { release_matrix0(m); m = NULL; } while (0)

#define is_attr_id(v)     (mode_id     && v == attr_id)
#define is_attr_eval(v)   (mode_eval   && v == attr_eval)
#define is_attr_ignore(v) (mode_ignore && attr_ignore[v])
#define is_attr_cont(v)   (attr_info[v].is_cont)
#define is_attr_disc(v)   (!(is_attr_id(v)     || is_attr_eval(v) ||  \
                             is_attr_ignore(v) || is_attr_cont(v)))
#define is_attr_aux(v)    (is_attr_id(v)     || is_attr_eval(v) ||  \
                           is_attr_ignore(v))

#define LGAMMA(x)  ((FLOAT_T)lngamma((double)(x)))
#define DIGAMMA(x) ((FLOAT_T)digamma((double)(x)))

typedef FLOAT_T **MATRIX;
typedef FLOAT_T *VECTOR;

typedef struct HASH_NODE {
  char *key;
  int id;
  int count;
  struct HASH_NODE *next;
} *HASH_NODE_PTR;

typedef struct _HASHTABLE {
  HASH_NODE_PTR *table;
  int size;
  int serial_id;
  int used;
} HASHTABLE;

typedef struct HASH_ENTRY {
  char *key;
  int id;
  int count;
} *HASH_ENTRY_PTR;

typedef struct FREQ {
  int *id;
  VECTOR *gval;
  int freq;
  int line_no;
  int case_id;
  int ans_id;
  int *miss_w;
  int **miss_g;
  int **obs_g;
  int num_miss_w;
  int *num_miss_g;
} *FREQ_PTR;

typedef struct GAUSS {
  int size;
  int *attr;
  char *str;
  struct GAUSS *next;
} *GAUSS_PTR;

typedef struct _ATTR_INFO {
  char is_cont;
  int gauss_id;
  int local_id;
} ATTR_INFO;

typedef struct STAT {
  /* For discrete attributes */
  int num_val;
  char *mode_val;
  int mode_count;

  /* For continuous attributes */
  FLOAT_T min;
  FLOAT_T max;
  FLOAT_T mean;
  FLOAT_T median;
  FLOAT_T var;
  FLOAT_T std;
  FLOAT_T p0;
  FLOAT_T p5;
  FLOAT_T p10;
  FLOAT_T p25;
  FLOAT_T p50;
  FLOAT_T p75;
  FLOAT_T p90;
  FLOAT_T p95;
  FLOAT_T p100;
} *STAT_PTR;

#ifndef M_POOL_SIZE
#  define M_POOL_SIZE 8388608 /* 8MB */
#endif
#ifndef V_POOL_SIZE
#  define V_POOL_SIZE 8388608 /* 8MB */
#endif

typedef long HEADER_ALIGN;

typedef union _M_HEADER {
  struct {
    union _M_HEADER *next;
    int dim0;
    int dim1;
  } info;
  HEADER_ALIGN dummy;
} M_HEADER;

typedef union _V_HEADER {
  struct {
    union _V_HEADER *next;
    int dim;
  } info;
  HEADER_ALIGN dummy;
} V_HEADER;

enum nbc_option_id {  /* [Note] Assigned integer values should be positive */
  O_N_CAND=1, O_AUTOCLASS, O_DAEM, O_ITEMP, O_SMOOTH_C,
  O_MIN_NEW_C_PARAM, O_DISSIM, O_MEM_PROB, O_EPSILON,
  O_EVAL, O_TARGET, O_NORM_JMERGE, O_SCORE, O_N_JSPLIT,
  O_HELP, O_READ_HPARAM, O_INPUT, O_INIT_C, O_OUTPUT,
  O_INIT_A, O_N_CLASS, O_SMOOTH_MEAN, O_MAX_ITER_RESTART, O_SMOOTH_VAR,
  O_MAX_ITER, O_SAVE_MEMORY, O_RESTART, O_MIN_SM_IMP, O_N_SM,
  O_ALPHA, O_CODE_IN, O_CODE_OUT, O_SEED,
  O_READ_PARAM, O_SUFFIX, O_SAVE_MESSAGE, O_TAU,
  O_GAUSS, O_MIN_C_PARAM, O_VERBOSE, O_SMOOTH_A,
  O_N_ATTR, O_OCC_ORDER, O_RANK, O_LOG_SCALE, O_ID,
  O_IGNORE, O_XHELP, O_HIDE_ATTR, O_RANK_HIDE,
  O_PREC_OUT, O_PREC_MSG, O_UNCOMPRESS, O_W_PURITY, O_W_WALLANCE,
  O_SHOW_ATTR_MEMP, O_SHOW_ATTR_CLUSTER, O_VERSION, O_REGULAR_STAT,
  O_NOISE_RATIO, O_NOISE_RATIO_GAUSS, O_EXTREME,
  O_NOOP
};
enum info_criterion_type {BIC=1, AUTOCLASS, VB};
enum cluster_type        {CL_MP=1};
enum init_type_for_class {RANDOM_C=1, NOISY_UNIFORM_C};
enum init_type_for_word  {RANDOM_W=1, NOISY_UNIFORM_W, WORD_FREQ};
enum relevance_measure   {REL_MP=1};
enum smooth_var_type     {SMOOTH_VAR_SAMPLE=1, SMOOTH_VAR_CONST};
enum smooth_auto_type    {MODE_SMOOTH_NONE=1, MODE_SMOOTH_AUTO};
enum file_format         {FMT_CSV=1, FMT_TSV};

void determine_format(char *,char *);
void open_log(void);
void parse_num_class(char *);
void parse_itemp(char *);
int parse_smooth_w(char *);
FLOAT_T parse_smooth_c(char *);
FLOAT_T parse_smooth_tau(char *);
void parse_smooth_s(char *);
void parse_extreme(char *);
void parse_rank(char *);
int parse_int(char *);
unsigned int parse_uint(char *);
FLOAT_T parse_float(char *);
void initialize_buffer(void);
#ifdef USE_ICONV
void initialize_iconv(void);
void initialize_iconv_sys(void);
#endif
void initialize_rand(void);
void set_smooth_auto(void);
void build_smooth_nominal(int);
void check_read_hparam(void);
void add_attr_ignore(char *);
void check_tuple_size(void);

void add_gauss(char *);
void build_attr_info(void);
void parse_gauss(void);
void initialize_hparam_gauss(void);

void em_main(void);
void no_em_main(void);
void em_iterate(int);
int em_iterate_body(int,int,int,FLOAT_T);
void compute_incomplete_prob(void);
FLOAT_T compute_loglike(void);
void initialize_param(void);
void compute_complete_prob(void);
void compute_complete_prob_daem(void);
void compute_conditional_prob(void);
FLOAT_T compute_log_prior(void);
void compute_expect(void);
void compute_expect_gauss(void);
void compute_expect_gauss0(int,int,FLOAT_T);
void compute_expect_gauss0_daem(int,int,FLOAT_T);
void compute_expect_gauss0_nodaem(int,int,FLOAT_T);
void compute_case_member_prob(void);
void compute_attr_member_prob(void);
void update_param(void);
int is_decreased(int, int, int);
int is_converged(int, int, int);
int is_converged_daem(int, int, int, int, FLOAT_T);
void save_log_itemp_best(int);
void restore_log_itemp_best(int);
void save_param_best(void);
void restore_param_best(void);
void save_param_opt(void);
void restore_param_opt(void);
void print_dot_message(int);
void print_daem_message(int,FLOAT_T);
void print_daem_message_sm(int,int,FLOAT_T);
FLOAT_T get_tau(int);
#ifdef VNBC
FLOAT_T get_alpha_w(int);
#else
FLOAT_T get_alpha_w(int,int,int);
#endif
FLOAT_T get_hparam_c(int);
FLOAT_T get_hparam_w(int,int,int);

void smem_iterate(int);
void compute_jmerge(void);
void sort_candidate(void);
void compute_incomplete_prob_part(int *);
void compute_conditional_prob_part(int *);
void save_incomplete_prob_part(int *);
void save_conditional_prob_part(int *);
void save_loglike_best_sm(int,int);
void restore_loglike_best_sm(int,int);
void save_log_itemp_best_sm(int,int);
void restore_log_itemp_best_sm(int,int);
void compute_jsplit(void);
void initialize_param_part(int *);
void compute_complete_prob_part(int *);
void compute_expect_part(int *);
void check_expect_part(int *);
int is_decreased_sm(int,int,int);
int is_converged_sm(int,int,int);
int is_converged_sm_daem(int,int,int,int,FLOAT_T);
void update_param_part(int *);
void check_param_part(int *);
void save_param_part(int *);
void restore_param_sm(void);
void print_decrease_sm(FILE *,int,int,FLOAT_T,FLOAT_T,FLOAT_T);
void alloc_param_sm();
void release_param_sm();
void save_param_sm();
void restore_param_sm();

FLOAT_T compute_bic(void);
FLOAT_T compute_cs(void);

void cluster_main(void);
void rank_main(void);
void eval_main(void);
void dissim_main(void);

void vbem_main(void);
void no_vbem_main(void);

void read_freq_body(void);
void read_freq(void);
int compare_word(const int *,const int *);
void compute_stats(void);

void expand_read_param_c(int,int);
void expand_read_param_w(int,int,int *,int *);
void expand_read_param_u(int,int);
void expand_read_param_s(int,int);
void read_param(void);
void read_hparam(void);

void expand_read_hparam_c(int,int);
void expand_read_hparam_w(int,int,int *,int *);
void expand_read_hparam_u(int,int);
void expand_read_hparam_s(int,int);

void print_version(FILE *);
void print_usage(FILE *);
void print_usage_full(FILE *);
void flush_desc(FILE *);
void flush_data(FILE *);
void print_data(FILE *,char *,...);
void initialize_printer(void);
void release_printer(void);
void add_desc_id(char *,...);
void add_desc_col(char *,...);
void add_desc_comm(char *,...);
void add_data(char *,...);
void add_data_str(char *);
void add_data_int(int);
void add_data_float(FLOAT_T);
char *get_case_id(int);
char *get_ans_class(int);
char *freq_entry_val(int, int);
char *mode_str(int);

void print_smooth(FILE *,int);
void print_loglike(FILE *,int,int);
void print_ic(FILE *);
void print_seed(FILE *);
void print_decrease(FILE *,int,int,FLOAT_T,FLOAT_T,FLOAT_T);
void print_jmerge(FILE *);
void print_jsplit(FILE *);
void print_loglike_sm(FILE *,int);
void print_num_iterate_sm(FILE *,int);
void print_num_result_sm(FILE *,int);
void print_time(FILE *);
void print_settings(FILE *);
void print_model(FILE *);
void print_basic_stats(FILE *);
void print_sample_stats(FILE *);
void print_regular_stat(FILE *);
void print_attr_list(FILE *);
void print_attr_freq_list(FILE *);
void print_case_freq_list(FILE *);
void print_ans_class_list(FILE *);
void print_size(FILE *);
void print_param(FILE *);
void print_hparam(FILE *);
void print_complete_prob(FILE *);
void print_incomplete_prob(FILE *);
void print_conditional_prob(FILE *);
void print_expectation(FILE *);
void print_case_member_prob(FILE *);
void print_attr_member_prob(FILE *);
void print_cluster_ml(FILE *);
void print_ml_class_freq(FILE *);
void print_rank(FILE *);
void print_cluster(FILE *);
void print_cluster_freq(FILE *);
void print_dissim_case_header(FILE *);
void print_dissim_case(FILE *,int,int,FLOAT_T);
void print_dissim_w_header(FILE *,int);
void print_dissim_w(FILE *,int,int,int,int,FLOAT_T);
void print_dissim_g_header(FILE *,int);
void print_dissim_g(FILE *,int,int,int,FLOAT_T);
void print_eval(FILE *);

void release_buffer(void);
void alloc_param(void);
void alloc_param_best(void);
void alloc_hparam(void);
void alloc_hparam_gauss(void);
void alloc_hparam_gauss_aux(void);
void alloc_expect(void);
void alloc_prob_mem(void);
void alloc_loglike(void);
void alloc_log_itemp(void);
void alloc_prob(void);
void alloc_ic(int);
void alloc_seed(void);
void alloc_ll_sm_size(void);
void alloc_loglike_sm(void);
void alloc_log_itemp_sm(void);
void alloc_jmerge(void);
void alloc_jsplit(void);
void alloc_candidate(void);
void alloc_prob_part(void);
void alloc_num_iterate_sm(void);
void alloc_num_result_sm(void);
void alloc_param_part(void);
void alloc_cluster(void);
void release_cluster(void);
void alloc_cluster_ml(void);
void alloc_rank(void);
void alloc_read_param(int,int *);
void alloc_read_hparam(int,int *);
void expand_loglike_all(void);
void expand_loglike_sm_all(int);
void alloc_eval(void);
void release_param(void);
void release_param_best(void);
void release_param_opt(void);
void release_hparam(void);
void release_hparam_gauss(void);
void release_expect(void);
void release_prob_mem(void);
void release_prob(void);
void release_loglike(void);
void release_log_itemp(void);
void release_ic(void);
void release_seed(void);
void release_jmerge(void);
void release_jsplit(void);
void release_candidate(void);
void release_prob_part(void);
void release_ll_sm_size(void);
void release_loglike_sm(void);
void release_log_itemp_sm(void);
void release_num_iterate_sm(void);
void release_num_result_sm(void);
void release_param_part(void);
void release_param_sm(void);
void release_cluster_ml(void);
void release_rank(void);
void release_rank_sm(void);
void release_read_param(int);
void release_read_hparam(int);
void release_eval(void);

void init_hashtable(HASHTABLE *);
void release_hashtable(HASHTABLE *);
int get_cardinality(HASHTABLE *);
int lookup_hashtable(HASHTABLE *, char *);
int register_hashtable(HASHTABLE *, char *);
char **get_keys(HASHTABLE *);
void get_mode_key(HASHTABLE *,char **,int *);
HASH_ENTRY_PTR get_entries(HASHTABLE *);
HASH_ENTRY_PTR get_entries_sorted_by_key(HASHTABLE *);
HASH_ENTRY_PTR get_entries_sorted_by_count(HASHTABLE *);

FLOAT_T current_sec(void);
FLOAT_T elapsed_sec(void);
void die(char *,...);
void die_line(char *,int,int,char *,char *,int);
void die_alloc(char *,int);
void die_realloc(char *,int);
void die_numeric(const char *,char *);
void die_dimension(const char *);
void print_msg(char *,...);
void print_msg_matrix(MATRIX,char *,...);
void print_msg_vector(VECTOR,char *,...);
char *conv_crlf(char *);
char *conv_base(char *,char *);
char *conv_sys(char *);
char *conv_msg(char *);
char *conv_delim(char *);
char *conv_out(char *);
char *conv_desc(char *);
char *conv_data(char *);
void conv_printf(char *,...);
void conv_fprintf(FILE *,char *,...);
void conv_sprintf(char *, char *, ...);
void conv_snprintf(char *, size_t, char *, ...);
void toupper_str(char *);
void tolower_str(char *);
void *xmalloc0(size_t,char *,int);
void *xrealloc0(void *,size_t,char *,int);
char *alloc_str0(int,char *,int);
char *dup_str0(char *,char *,int);
char *int2str(int);
char *int2strdup(int);
char *float2str(FLOAT_T);
char *float2strdup(FLOAT_T);
int xvsprintf(char *,const char *,va_list);
int xvsnprintf(char *,size_t,const char *,va_list);
int xsprintf(char *,const char *,...);
int xsnprintf(char *,size_t,const char *,...);

char **alloc_field(int);
void release_field(char **,int);
int split(int,const char *,char **,int *,const char *,const char *);
int split_comma(int,const char *,char **,int *);
int split_tab(int,const char *,char **,int *);

double lngamma(double);
double digamma(double);

FLOAT_T determinant(MATRIX);
FLOAT_T determinant_c(MATRIX);
FLOAT_T trace(MATRIX);
void set_diag(MATRIX, FLOAT_T);
void set_diag_matrix(MATRIX,FLOAT_T);
FLOAT_T lu(MATRIX,MATRIX,MATRIX,int *);
FLOAT_T cholesky(MATRIX,MATRIX);
void invert_u(MATRIX,MATRIX);
void invert_l(MATRIX,MATRIX);
void invert_u_p(MATRIX,MATRIX,int *);
void invert_l_p(MATRIX,MATRIX,int *);
FLOAT_T invert(MATRIX,MATRIX);
FLOAT_T invert_c(MATRIX,MATRIX);
FLOAT_T qform(MATRIX,VECTOR);
void transpose(MATRIX,MATRIX);
void sum_matrix(MATRIX,MATRIX,MATRIX);
void diff_matrix(MATRIX,MATRIX,MATRIX);
void prod_matrix(MATRIX,MATRIX,MATRIX);
void prod_matrix_vector(MATRIX,VECTOR,VECTOR);
void sum_vector(VECTOR,VECTOR,VECTOR);
void diff_vector(VECTOR,VECTOR,VECTOR);
void proj_matrix(MATRIX,int *,int *,MATRIX);
void proj_vector(VECTOR,int *,VECTOR);
FLOAT_T prod_vector(VECTOR,VECTOR);
void mul_const_matrix(MATRIX,FLOAT_T);
void mul_const_vector(VECTOR,FLOAT_T);
void div_const_matrix(MATRIX,FLOAT_T);
void div_const_vector(VECTOR,FLOAT_T);
void fprint_vector(FILE *,VECTOR);
void fprint_matrix(FILE *,MATRIX);
void fprint_matrix_p(FILE *,MATRIX,int *);
void print_vector(VECTOR);
void print_matrix(MATRIX);
void print_matrix_p(MATRIX,int *);
void matrix2vector(MATRIX,VECTOR);
void vector2matrix(VECTOR,MATRIX);
int verify_matrix(MATRIX);
int verify_vector(VECTOR);

void initialize_matrix_pool(void);
void initialize_vector_pool(void);
void release_matrix_pool(void);
void release_vector_pool(void);
MATRIX alloc_matrix(int,int);
VECTOR alloc_vector(int);
void release_matrix0(MATRIX);
void release_vector0(VECTOR);

void set_matrix(MATRIX,FLOAT_T **);
void set_vector(VECTOR,FLOAT_T *);
MATRIX alloc_set_matrix(int,int,FLOAT_T **);
VECTOR alloc_set_vector(int,FLOAT_T *);
MATRIX alloc_clear_matrix(int,int);
VECTOR alloc_clear_vector(int);
void copy_matrix(MATRIX,MATRIX);
void copy_vector(VECTOR,VECTOR);
MATRIX dup_matrix(MATRIX);
VECTOR dup_vector(VECTOR);
void set_vprod_matrix(MATRIX,VECTOR,VECTOR);
MATRIX alloc_vprod_matrix(VECTOR,VECTOR);
VECTOR alloc_proj_vector(VECTOR,int *);
MATRIX alloc_proj_matrix(MATRIX,int *,int *);
VECTOR alloc_proj_vector_n(VECTOR,int,int *);
MATRIX alloc_proj_matrix_n(MATRIX,int,int,int *,int *);
void copy_upper_to_lower(MATRIX);
void clear_matrix(MATRIX);
void clear_vector(VECTOR);
#ifdef DISABLE_DIM
int fdim0(MATRIX);
int fdim1(MATRIX);
int fdim(VECTOR);
#  define dim0(m) (fdim0(m))
#  define dim1(m) (fdim1(m))
#  define dim(v)  (fdim(v))
#else
#  define dim0(m) (((M_HEADER *)(m) - 1)->info.dim0)
#  define dim1(m) (((M_HEADER *)(m) - 1)->info.dim1)
#  define dim(v)  (((V_HEADER *)(v) - 1)->info.dim)
#endif

FLOAT_T gauss(VECTOR,VECTOR,MATRIX);
FLOAT_T gauss_marginal(VECTOR,VECTOR,MATRIX,int *);
FLOAT_T gauss_prec(VECTOR,VECTOR,MATRIX,MATRIX,FLOAT_T);
FLOAT_T gauss_stat(FLOAT_T,VECTOR,MATRIX,VECTOR,MATRIX);
FLOAT_T gauss_stat_prec(FLOAT_T,VECTOR,MATRIX,VECTOR,MATRIX,MATRIX,FLOAT_T);
FLOAT_T gauss_cond(VECTOR,VECTOR,MATRIX,int *);
FLOAT_T lgauss(VECTOR,VECTOR,MATRIX);
FLOAT_T lgauss_marginal(VECTOR,VECTOR,MATRIX,int *);
FLOAT_T lgauss_prec(VECTOR,VECTOR,MATRIX,MATRIX,FLOAT_T);
FLOAT_T lgauss_stat(FLOAT_T,VECTOR,MATRIX,VECTOR,MATRIX);
FLOAT_T lgauss_stat_prec(FLOAT_T,VECTOR,MATRIX,VECTOR,MATRIX,MATRIX,FLOAT_T);
FLOAT_T lgauss_cond(VECTOR,VECTOR,MATRIX,int *);

FLOAT_T wishart(MATRIX,MATRIX,FLOAT_T);
FLOAT_T lwishart(MATRIX,MATRIX,FLOAT_T);
FLOAT_T lwishart_det(MATRIX,MATRIX,MATRIX,FLOAT_T,FLOAT_T,FLOAT_T);
FLOAT_T lwishart_coeff(int,FLOAT_T);

FLOAT_T lgamma_d_coeff(FLOAT_T);
FLOAT_T lgamma_d(FLOAT_T,FLOAT_T,FLOAT_T);
FLOAT_T gamma_d(FLOAT_T,FLOAT_T,FLOAT_T);

FLOAT_T lstudents_t(VECTOR,VECTOR,MATRIX,FLOAT_T);
FLOAT_T students_t(VECTOR,VECTOR,MATRIX,FLOAT_T);
FLOAT_T lstudents_t_marginal(VECTOR,VECTOR,MATRIX,FLOAT_T,int *);
FLOAT_T students_t_marginal(VECTOR,VECTOR,MATRIX,FLOAT_T,int *);

/* random.c (mt19937ar.c) */
void init_genrand(unsigned long);
long genrand_int31(void);
double genrand_res53(void);
#define random_seed(s) init_genrand(s)
#define random_int     genrand_int31
#define random_float   genrand_res53

FLOAT_T get_quantile(FLOAT_T,int,FLOAT_T *);
FLOAT_T get_quantile_count(FLOAT_T,int,FLOAT_T *,int *);

#endif /* NBC_DEF_H */

