/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

char version[] =
#ifdef _OPENMP
  TOOL_NAME " version " VERSION " (VB, built on " __DATE__ ", with OpenMP)\n";
#else
  TOOL_NAME " version " VERSION " (VB, built on " __DATE__ ")\n";
#endif

char usage_full[] =
  "[Usage] vnbc -f basename -k #classes -x #attributes [option...]\n"
  "Flags with an argument:\n"
  " --target   (-f) Prefix of input/output files                      [Required]\n"
  " --suffix   (-s) Suffix of output files                           [No suffix]\n"
  " --n-class  (-k) Number of classes or Kmin:Kmax:Kstep\n"
  "                                                [Required unless -R/-H given]\n"
  " --n-attr   (-x) Number of attributes                              [Required]\n"
  " --gauss    (-u) Indices of continuous attributes        [No continous attrs]\n"
  " --id       (-I) Index of the ID attribute\n"
  " --ignore   (-y) Indices of the attributes to be ignored\n"
  " --eval     (-E) Index of the answer class attribute (Evaluation enabled)\n"
  " --input    (-i) File extension of the input files\n"
  "                 (csv:comma-separated, txt:tab-separated)       [Auto-detect]\n"
  " --output   (-j) File extension of the output files\n"
  "                 (csv:comma-separated, txt:tab-separated)    [Same as inputs]\n"
  " --smooth-c (-c) Initial hyperparemters for the class attribute\n"
  "                 (auto:AutoClass, other positive float: as-is) [No smoothing]\n"
  " --smooth-a (-w) Initial hyperparameters for discrete attributes\n"
  "                 (auto:AutoClass, other positive float: as-is) [No smoothing]\n"
  " --smooth-mean (-K) Initial hyperparameters for means\n"
  " --smooth-var  (-L) Initial hyperparameters for variances\n"
  "                    (sample:<val>, const:<val>, or other short-forms)\n"
  "                                                                 [sample:1.0]\n"
  " --tau      (-T) Tau in the Normal-Wishart priors\n"
  " --alpha    (-O) Alpha in the Normal-Wishart priors\n"
  " --epsilon  (-e) Threshold in judging convergence                    [1.0e-6]\n"
  " --max-iterate (-m) Max num. of EM iterations  [Iterate EM until convergence]\n"
  " --seed     (-r) Random seed                       [Returned value of time()]\n"
  " --restart  (-n) Number of restarts                              [No restart]\n"
  " --max-iter-restart (-l)\n"
  "                 Max num. of preliminary VBEM iterations in random restart\n"
  "                                             [Iterate VBEM until convergence]\n"
  " --noise         Noises in initialization of categorical hyperparams\n"
  " --noise-gauss   Noises in initialization of hyperparams of Gaussian means\n"
  " --extreme       Threshold for judging extreme values                  [0.05]\n"
  " --rank  (-z)    Number of attribute values highly relevant to a class,\n"
  "                 or threshold for the relevance (Ranking is enabled)\n"
  "                                                        [Ranking is disabled]\n"
  " --itemp (-B) Binit:Brate\n"
  "                 where Binit is the initial value of the inverse temperature,\n"
  "                 and Brate is the update rate of the inverse temperature\n"
  "                                                                    [0.1:1.5]\n"
  " --n-sm     (-o) Max num. of split-merge operations (SMEM enabled)\n"
  "                                                              [SMEM disabled]\n"
  " --n-cand   (-a) Max num. of SM candidates to be examined          [No limit]\n"
  " --n-jsplit (-G) Max num. of split candidates to be examined       [No limit]\n"
  " --min-c-param\n"
  "              Lower limit of prob. of the class to be split/merged [No limit]\n"
  " --min-new-c-param\n"
  "              Lower limit of prob. of the splited/merged class     [No limit]\n"
  " --min-sm-imp Lower limit of improvement by spilt-merge oparations\n"
  "                                     [Threshold of convergence (given by -e)]\n"
#ifdef USE_ICONV
#ifdef CHARCODE
  " --code-in  (-p) Input character code   [" CHARCODE "]\n"
  " --code-out (-q) Output character code  [" CHARCODE "]\n"
  "                   (run 'iconv -l' for available code names)\n"
#else
  " --code-in  (-p) Input character code   [" CHARCODE_INTERNAL "]\n"
  " --code-out (-q) Output character code  [" CHARCODE_INTERNAL "]\n"
  "                 (run 'iconv -l' for available code names)\n"
#endif
#endif
  " --prec-out  Number of significant digits of\n"
  "               a floating point number in the output files               [12]\n"
  " --prec-msg  Number of digits after the decimal point of\n"
  "               a floating point number in the terminal messages           [6]\n"
  " Flags with no argument:\n"
  " --read-hparam    (-H) Disable VBEM and read hyperparameters from a file\n"
  " --mem-prob       (-D) Compute and output the membership distributions\n"
  " --predict        (-P) Compute and output the predictive distributions\n"
  " --dissim              Compute and output the dissimilarities\n"
  " --uncompress          Uncompress the cases having the same attribute values\n"
  " --verbose             Output the temporarily computed results to log file\n"
  " --daem           (-b) Enable DAEM\n"
  " --norm-jmerge         Normalize Jmerge\n"
  " --regular-stat        Ignore extreme values in getting sample means/covars\n"
  " --occ-order           Print attvals in the appearance order     [Lex. order]\n"
  " --hide-attr      (-V) Hide attribute values in the output files\n"
  " --hide-rank           Hide irrevant attribute values in ranking\n"
  " --show-attr-memp      Show the membership probabilities for attributes\n"
  " --show-attr-cluster   Show the clusters of attribute values\n"
  " --save-msg       (-S) Output the terminal messages to the files  [To stderr]\n"
  " --version        (-v) Print the version number\n"
  " --help           (-h) Print brief descriptions on major flags\n"
  " --helpall        (-X) Print brief descriptions on all flags\n";

char usage[] =
  "[Usage] vnbc -f basename -k #classes -x #attributes [option...]\n"
  "Flags with an argument:\n"
  " --target   (-f) Prefix of input/output files                      [Required]\n"
  " --suffix   (-s) Suffix of output files                           [No suffix]\n"
  " --n-class  (-k) Number of classes or Kmin:Kmax:Kstep\n"
  "                                                [Required unless -R/-H given]\n"
  " --n-attr   (-x) Number of attributes                              [Required]\n"
  " --gauss    (-u) Indices of continuous attributes        [No continous attrs]\n"
  " --id       (-I) Index of the ID attribute\n"
  " --ignore   (-y) Indices of the attributes to be ignored\n"
  " --eval     (-E) Index of the answer class attribute (Evaluation enabled)\n"
  " --output   (-j) File extension of the output files\n"
  "                 (csv:comma-separated, txt:tab-separated)    [Same as inputs]\n"
  " --epsilon  (-e) Threshold in judging convergence                    [1.0e-6]\n"
  " --max-iter (-m) Max num. of EM iterations     [Iterate EM until convergence]\n"
  " --restart  (-n) Number of restarts                              [No restart]\n"
  " --rank     (-z) Number of attribute values highly relevant to a class,\n"
  "                 or threshold for the relevance (Ranking enabled)\n"
  "                                                           [Ranking disabled]\n"
#ifdef USE_ICONV
#ifdef CHARCODE
  " --code-in  (-p) Input character charcode    [" CHARCODE "]\n"
  " --code-out (-q) Output character charcode   [" CHARCODE "]\n"
  "                 (run 'iconv -l' for available charcodes)\n"
#else
  " --code-in  (-p) Input character charcode    [" CHARCODE_INTERNAL "]\n"
  " --code-out (-q) Output character charcode   [" CHARCODE_INTERNAL "]\n"
  "                 (run 'iconv -l' for available charcodes)\n"
#endif
#endif
  "Flags with no argument:\n"
  " --read-hparam (-H) Disable VBEM and read hyperparameters from a file\n"
  " --mem-prob    (-D) Compute and output the membership distributions\n"
  " --predict     (-P) Compute and output the predictive distributions\n"
  " --save-msg    (-S) Output the terminal messages to the files  [To stderr]\n"
  " --version     (-v) Print the version number\n"
  " --help        (-h) Print brief descriptions on major flags\n"
  " --helpall     (-X) Print brief descriptions on all flags\n";

void print_seed(FILE *fp)
{
  int k,u,r;

  if (mode_search_num_class) {
    u = 0;
    for (k = kmin; k <= kmax; k += kstep) {
      add_desc_id("SEED-K%d",k);
      add_desc_col("Num. of classes");
      add_desc_col("Restart");
      add_desc_col("Random seed");
      add_desc_col("Num. of updates");
      add_desc_col("Variational free energy");
      add_desc_comm("Best seed = %d",log_r_best[u]);
      flush_desc(fp);

      for (r = 0; r < num_init; r++) {
        if (mode_set_seed)
          print_data(fp,"%d,%d,N/A,%d,%.Ng",
                     k,r,log_num_iterate[u][r],log_final_free_energy[u][r]);
        else
          print_data(fp,"%d,%d,%u,%d,%.Ng",
                     k,r,log_seed[u][r],log_num_iterate[u][r],
                     log_final_free_energy[u][r]);
      }
      u++;
    }

  }
  else {
    add_desc_id("SEED");
    add_desc_col("Num. of classes");
    add_desc_col("Restart");
    add_desc_col("Random seed");
    add_desc_col("Num. of updates");
    add_desc_col("Variational free energy");
    add_desc_comm("Best seed = %d",log_r_best[0]);
    flush_desc(fp);

    for (r = 0; r < num_init; r++) {
      if (mode_set_seed)
        print_data(fp,"%d,%d,N/A,%d,%.Ng",
                   num_class,r,log_num_iterate[0][r],
                   log_final_free_energy[0][r]);
      else
        print_data(fp,"%d,%d,%u,%d,%.Ng",
                   num_class,r,log_seed[0][r],
                   log_num_iterate[0][r],log_final_free_energy[0][r]);
    }
  }

  fflush(fp);
}

void print_settings(FILE *fp)
{
  add_desc_id("SETTING-VERSION");
  add_desc_col("Version");
  flush_desc(fp);
  fputs(VERSION,fp);

  add_desc_id("SETTING-FILE");
  add_desc_col("Input/Output");
  add_desc_col("Contents");
  add_desc_col("Filename");
  flush_desc(fp);
  print_data(fp,"Input,Cases,%s",fname_freq);
  if (mode_em)
    print_data(fp,"Output,Hyperparameters,%s",fname_hparam);
  else
    print_data(fp,"Input,Hyperparameters,%s",fname_hparam);
  if (mode_memp)    print_data(fp,"Output,Member distributions,%s",fname_memp);
  if (mode_predict) print_data(fp,"Output,Predictive distributions,%s",fname_predict);
  if (mode_dissim)  print_data(fp,"Output,Dissimilarities,%s",fname_dissim);
  if (mode_cluster) print_data(fp,"Output,Clusters,%s",fname_cluster);
  if (mode_rank)    print_data(fp,"Output,Rankings,%s",fname_rank);
  if (mode_eval)    print_data(fp,"Output,Evaluation results,%s",fname_eval);
  print_data(fp,"Output,Logs,%s",fname_log);
  if (mode_msg)     print_data(fp,"Output,Message logs,%s",fname_msg);

  add_desc_id("SETTING-FUNCTION");
  add_desc_col("Major function");
  add_desc_col("Enabled/Disabled");
  flush_desc(fp);
  print_data(fp,"Reading hyperparameters,%s",mode_str(!mode_em));
  print_data(fp,"Member distributions,%s",mode_str(mode_memp));
  print_data(fp,"Predictive distributions,%s",mode_str(mode_predict));
  print_data(fp,"Clustering,%s",mode_str(mode_cluster));
  print_data(fp,"Relevance analysis,%s",mode_str(mode_rank));
  print_data(fp,"Dissimilarities,%s",mode_str(mode_dissim));
  print_data(fp,"Model selection,%s",mode_str(mode_search_num_class));
  print_data(fp,"Deterministic annealing EM,%s",mode_str(mode_daem));
  if (num_sm > 0)
    print_data(fp,"Split-merge EM,%s",mode_str(1));
  else
    print_data(fp,"Split-merge EM,%s",mode_str(0));
  print_data(fp,"Evaluation,%s",mode_str(mode_eval));

  print_model(fp);

  if (mode_em) {
    add_desc_id("SETTING-EM");
    add_desc_col("Control parameter");
    add_desc_col("Value");
    flush_desc(fp);
    print_data(fp,"Random seed,%d",rand_seed);
    print_data(fp,"Threshold for convergence,%.Ng",epsilon);
    print_data(fp,"Max num Of EM iterations,%d",max_iterate);
    if (num_init > 1) {
      print_data(fp,"Random restarts,%s",mode_str(1));
      print_data(fp,"Num. of restarts,%d",num_init);
      print_data(fp,"Max num of EM iterations in restarts,%d",max_init_iterate);
    }
    else print_data(fp,"Random restarts,%s",mode_str(0));
    print_data(fp,"Noises in initialization of "
               "categorical hyperparameters,%.Ng",
               noise_ratio);
    print_data(fp,"Noises in initialization of "
               "hyperparameters of Gaussian means,%.Ng",
               noise_ratio_gauss);
    print_data(fp,"Threshold for extreme values (Min.),%.Ng",q_regular_min);
    print_data(fp,"Threshold for extreme values (Max.),%.Ng",q_regular_max);
  }

  if (mode_search_num_class) {
    add_desc_id("SETTING-SELECT");
    add_desc_col("Control parameter");
    add_desc_col("Value");
    flush_desc(fp);
    print_data(fp,"Model score,Variational free energy");
  }

  if (mode_rank) {
    add_desc_id("SETTING-RANK");
    add_desc_col("Control parameter");
    add_desc_col("Value");
    flush_desc(fp);
    print_data(fp,"Hiding less relevant cases/attvals,%s",
               mode_str(mode_rank_hide));
    if (mode_rank_th)
      print_data(fp,"Threshold for relevant cases/attvals,%.Ng",rank_th);
    else
      print_data(fp,"Num of relevant cases/attvals,%d",num_rank);
  }

  if (mode_daem) {
    add_desc_id("SETTING-DAEM");
    add_desc_col("Control parameter");
    add_desc_col("Value");
    flush_desc(fp);
    print_data(fp,"Initial value of inverse temperature,%.Ng",itemp_init);
    print_data(fp,"Update rate of inverse temperature,%.Ng",itemp_rate);
  }

  if (num_sm > 0) {
    add_desc_id("SETTING-SMEM");
    add_desc_col("Control parameter");
    add_desc_col("Value");
    flush_desc(fp);
    print_data(fp,"Num of split-merge operations,%d",num_sm);
    print_data(fp,"Max num of split-merge candidates,%d",max_candidate);
    print_data(fp,"Max num of split candidates,%d",max_candidate);
    print_data(fp,"Normalization of J-merge,%s",mode_str(mode_norm));
    print_data(fp,"Min class prob. before split-merge operation"
               " [0: No limit],%.Ng",min_param_c);
    print_data(fp,"Min class prob. after split-merge operation"
               " [0: No limit],%.Ng",min_new_param_c);
    print_data(fp,"Min improvement,%.Ng",min_smem_improve);
  }

  if (mode_eval) {
    add_desc_id("SETTING-EVAL");
    add_desc_col("Control parameter");
    add_desc_col("Value");
    flush_desc(fp);
    print_data(fp,"Weight in F-measure of Purity,%.Ng",weight_purity);
    print_data(fp,"Weight in F-measure of Wallance's measures,%.Ng",
               weight_wallance);
  }

  add_desc_id("SETTING-INOUT");
  add_desc_col("Control parameter");
  add_desc_col("Value");
  flush_desc(fp);
  print_data(fp,"Target,%s",fname_base);
  if (mode_suffix)
    print_data(fp,"Suffix,%s",fname_suffix);
  else
    print_data(fp,"Suffix,(none)");
  print_data(fp,"Input file extension,%s",
             ((format_in == FMT_CSV) ?  FEXT_CSV : FEXT_TSV));
  print_data(fp,"Output file extension,%s",
             ((format_out == FMT_CSV) ? FEXT_CSV : FEXT_TSV));
#ifdef USE_ICONV
  print_data(fp,"Input charcode,%s",charcode_in);
  print_data(fp,"Output charcode,%s",charcode_out);
#endif

  add_desc_id("SETTING-MISC");
  add_desc_col("Control parameter");
  add_desc_col("Value");
  flush_desc(fp);
  print_data(fp,"Ignoring extreme values when computing sample means/covars,%s",
             mode_str(mode_regular_stat));
  print_data(fp,"Writing attvals in lexicographical order,%s",
             mode_str(mode_lex_order));
  print_data(fp,"Uncompressing identical objects,%s",
             mode_str(mode_uncompress));
  print_data(fp,"Saving messages,%s",mode_str(mode_msg));
  print_data(fp,"Being verbose,%s",mode_str(mode_verbose));
  print_data(fp,"Showing attvals,%s",mode_str(!mode_hide_attr));
  print_data(fp,"Showing membership probs. for attributes,%s",
             mode_str(mode_show_attr_memp));
  print_data(fp,"Showing clusters of attvals,%s",
             mode_str(mode_show_attr_cluster));
#ifdef _OPENMP
  if (omp_get_thread_num() == 0) {
    print_data(fp,"Max num of threads,%d",omp_get_max_threads());
  }
#endif
  print_data(fp,"Precision of float-point numbers [file output],%d",prec_out);
  print_data(fp,"Precision of float-point numbers [terminal message],%d",
             prec_msg);
#ifndef NO_LDOUBLE
  print_data(fp,"Byte size [double],%d",sizeof(double));
  print_data(fp,"Byte size [long-double],%d",sizeof(FLOAT_T));
#endif

  fflush(fp);
}

void print_model(FILE *fp)
{
  int v,v_d,u,n,a;

  add_desc_id("SETTING-MODEL-BASE");
  add_desc_col("Model specification");
  flush_desc(fp);
  if (mode_search_num_class) {
    print_data(fp,"Num of classes [min],%d",kmin);
    print_data(fp,"Num of classes [max],%d",kmax);
    print_data(fp,"Num of classes [step],%d",kstep);
  }
  else {
    print_data(fp,"Num of classes,%d",num_class);
  }
  print_data(fp,"Num of attributes,%d",tuple_size);
  print_data(fp,"Num of discrete attributes,%d",tuple_size_d);
  print_data(fp,"Num of continuous attributes,%d",tuple_size_c);

  if (mode_id)
    print_data(fp,"ID attribute,%d",attr_id);
  else
    print_data(fp,"ID attribute,(none)");

  if (mode_eval)
    print_data(fp,"Answer class attribute,%d",attr_eval);
  else
    print_data(fp,"Answer class attribute,(none)");

  if (mode_ignore) {
    add_data_str("Ignored attributes");
    for (v=0; v<tuple_size; v++)
      if (is_attr_ignore(v)) add_data_int(v);
    flush_data(fp);
  }
  else print_data(fp,"Ignored attributes,(none)");

  if (mode_nominal) {
    add_data_str("Discrete attributes");
    for (v=0; v<tuple_size; v++)
      if (is_attr_disc(v)) add_data_int(v);
    flush_data(fp);
  }
  else print_data(fp,"Discrete attributes,(none)");

  if (mode_gauss) {
    add_data_str("Continuous attributes");
    for (v=0; v<tuple_size; v++)
      if (is_attr_cont(v)) add_data_int(v);
    flush_data(fp);

    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];
      add_data("Attributes in Gauss%d",u);
      for (a=0; a<n; a++)
        add_data_int(gauss_attr[u][a]);
      flush_data(fp);
    }
    for (u=0; u<gauss_size; u++)
      print_data(fp,"Dimension of Gauss%d,%d",u,gauss_attr_size[u]);
  }
  else print_data(fp,"Continuous attributes,(none)");

  if (mode_em) {
    add_desc_id("SETTING-MODEL-HPARAM");
    add_desc_col("Hyperparameter");
    add_desc_col("Value");
    flush_desc(fp);
    if (mode_smooth_auto_c == MODE_SMOOTH_AUTO)
      print_data(fp,"Pseudo count [class],AutoClass");
    else
      print_data(fp,"Pseudo count [class],%.Ng",smooth_c);
    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;
      v_d = attr_info[v].local_id;
      if (mode_smooth_auto_w[v_d] == MODE_SMOOTH_AUTO)
        print_data(fp,"Pseudo count [attr %d],AutoClass",v);
      else
        print_data(fp,"Pseudo count [attr %d],%.Ng",v,smooth_w[v_d]);
    }

    if (mode_smooth_auto_tau == MODE_SMOOTH_AUTO)
      print_data(fp,"Prior Tau,AutoClass");
    else
      print_data(fp,"Prior Tau,%.Ng",tau);

    if (alpha_w > 0)
      print_data(fp,"Prior Alpha,%.Ng",alpha_w);
    else
      print_data(fp,"Prior Alpha,Dimension + %.Ng",(FLOAT_T)ALPHA_W_NOISE);

    if (mode_smooth_u)
      print_data(fp,"Prior Mean,%.Ng",smooth_u);
    else
      print_data(fp,"Prior Mean,Sample mean");

    if (mode_smooth_s == SMOOTH_VAR_CONST)
      print_data(fp,
                 "Prior Sigma,"
                 "Diagonal with all elements (variances) being %.Ng",
                 smooth_s);
    else
      print_data(fp,
                 "Prior Sigma,"
                 "Sample covariances multiplied by %.Ng",
                 smooth_s);
  }

  fflush(fp);
}

void print_basic_stats(FILE *fp)
{
  int v;

  if (mode_nominal) {
    add_desc_id("STAT-BASIC-CAT");
    add_desc_col("Attr. ID");
    add_desc_col("Num. of distinct values");
    add_desc_col("Mode value");
    add_desc_col("Mode count");
    add_desc_comm("Only for discrete attributes");
    flush_desc(fp);

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;

      add_data_int(v);
      add_data_int(attr_stat[v].num_val);
      add_data_str(attr_stat[v].mode_val);
      add_data_int(attr_stat[v].mode_count);
      flush_data(fp);
    }
  }

  if (mode_gauss) {
    add_desc_id("STAT-BASIC-GAUSS");
    add_desc_col("Gauss ID");
    add_desc_col("Attr. ID");
    add_desc_col("Min.");
    add_desc_col("Mean");
    add_desc_col("Median");
    add_desc_col("Max.");
    add_desc_col("Var.");
    add_desc_col("Std.");
    add_desc_comm("Only for continuous attributes");
    flush_desc(fp);

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_cont(v)) continue;

      add_data_int(attr_info[v].gauss_id);
      add_data_int(v);
      add_data_float(attr_stat[v].min);
      add_data_float(attr_stat[v].mean);
      add_data_float(attr_stat[v].median);
      add_data_float(attr_stat[v].max);
      add_data_float(attr_stat[v].var);
      add_data_float(attr_stat[v].std);
      flush_data(fp);
    }
  }

  if (mode_gauss) {
    add_desc_id("STAT-PERCENTILE");
    add_desc_col("Attr. ID");
    add_desc_col("Percent");
    add_desc_col("Percentile");
    add_desc_comm("Only for continuous attributes");
    flush_desc(fp);

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_cont(v)) continue;

      print_data(fp,"%d,0,%.Ng",v,attr_stat[v].p0);
      print_data(fp,"%d,5,%.Ng",v,attr_stat[v].p5);
      print_data(fp,"%d,10,%.Ng",v,attr_stat[v].p10);
      print_data(fp,"%d,25,%.Ng",v,attr_stat[v].p25);
      print_data(fp,"%d,50,%.Ng",v,attr_stat[v].p50);
      print_data(fp,"%d,75,%.Ng",v,attr_stat[v].p75);
      print_data(fp,"%d,90,%.Ng",v,attr_stat[v].p90);
      print_data(fp,"%d,95,%.Ng",v,attr_stat[v].p95);
      print_data(fp,"%d,100,%.Ng",v,attr_stat[v].p100);
    }
  }

  fflush(fp);
}

void print_sample_stats(FILE *fp)
{
  int u,a,a0,a1;

  add_desc_id("STAT-GAUSS-MEAN");
  add_desc_col("Gauss ID");
  add_desc_col("Attr. ID");
  add_desc_col("Mean");
  flush_desc(fp);

  for (u=0; u<gauss_size; u++)
    for (a=0; a<gauss_attr_size[u]; a++)
      print_data(fp,"%d,%d,%.Ng",u,gauss_attr[u][a],sample_u_raw[u][a]);

  add_desc_id("STAT-GAUSS-COVAR");
  add_desc_col("Gauss ID");
  add_desc_col("Attr. ID");
  add_desc_col("Attr. ID");
  add_desc_col("(Co)Variance");
  flush_desc(fp);

  for (u=0; u<gauss_size; u++)
    for (a0=0; a0<gauss_attr_size[u]; a0++)
      for (a1=a0; a1<gauss_attr_size[u]; a1++)
        print_data(fp,"%d,%d,%d,%.Ng",
                   u,gauss_attr[u][a0],gauss_attr[u][a1],
                   sample_s_raw[u][a0][a1]);

  fflush(fp);
}

void print_regular_stat(FILE *fp)
{
  int u,a,a0,a1;

  add_desc_id("STAT-REGULAR-Q");
  add_desc_col("Q-min.");
  add_desc_col("Q-max.");
  add_desc_comm("The values below Q-min. or above Q-max. "
                "are considered as extreme values");
  flush_desc(fp);

  print_data(fp,"%.Ng,%.Ng",q_regular_min,q_regular_max);

  add_desc_id("STAT-REGULAR-MEAN");
  add_desc_col("Gauss ID");
  add_desc_col("Attr. ID");
  add_desc_col("Min.");
  add_desc_col("Mean");
  add_desc_col("Max.");
  add_desc_comm("These are the statistics when ignoring extreme values");
  flush_desc(fp);

  for (u=0; u<gauss_size; u++)
    for (a=0; a<gauss_attr_size[u]; a++)
      print_data(fp,"%d,%d,%.Ng,%.Ng,%.Ng",
                 u,gauss_attr[u][a],
                 regular_min[u][a],regular_u[u][a],regular_max[u][a]);

  add_desc_id("STAT-REGULAR-COVAR");
  add_desc_col("Gauss ID");
  add_desc_col("Attr. ID");
  add_desc_col("Attr. ID");
  add_desc_col("(Co)Variance");
  add_desc_comm("These are the statistics when ignoring extreme values");
  flush_desc(fp);

  for (u=0; u<gauss_size; u++)
    for (a0=0; a0<gauss_attr_size[u]; a0++)
      for (a1=a0; a1<gauss_attr_size[u]; a1++)
        print_data(fp,"%d,%d,%d,%.Ng",
                   u,gauss_attr[u][a0],gauss_attr[u][a1],regular_s[u][a0][a1]);

  fflush(fp);
}

void print_attr_list(FILE *fp)
{
  int v,v_d,i,j;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("ATTVAL%d",v);
    add_desc_col("ID for the attribute value");
    add_desc_col("Attr. value x%d",v);
    flush_desc(fp);

    for (i=0; i<num_word[v_d]; i++) {
      j = print_order[v_d][i];
      print_data(fp,"%d,%s",j,word_list[v_d][j]);
    }
  }

  fflush(fp);
}

void print_attr_freq_list(FILE *fp)
{
  int v,v_d,i,j;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("FREQ-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("ID of attr. value x%d",v);
    add_desc_col("Attr. value x%d",v);
    add_desc_col("#occurrences N(x%d)",v);
    flush_desc(fp);

    for (i=0; i<num_word[v_d]; i++) {
      j = print_order[v_d][i];
      print_data(fp,"%d,%d,%s,%d",v,j,word_list[v_d][j],word_freq_list[v_d][j]);
    }
  }

  fflush(fp);
}

void print_case_freq_list(FILE *fp)
{
  int v,e;

  add_desc_id("FREQ-CASE");
  add_desc_col("Case ID");
  if (mode_eval) add_desc_col("Answer class");
  if (!mode_hide_attr) add_desc_col("Attribute values x");
  add_desc_col("#occurrences N(x)");
  flush_desc(fp);

  for (e=0; e<num_entry; e++) {
    add_data_str(get_case_id(e));
    if (mode_eval) add_data_str(get_ans_class(e));
    if (!mode_hide_attr)
      for (v=0; v<tuple_size; v++)
        if (!is_attr_aux(v))
          add_data_str(freq_entry_val(e,v));
    add_data_int(freq_list[e].freq);
    flush_data(fp);
  }

  fflush(fp);
}

void print_size(FILE *fp)
{
  int k,v,v_d,v_c,u,n;
  int num_param_c,num_param_d,total_num_param;

  add_desc_id("DATASIZE");
  add_desc_col("Total num. of cases");
  add_desc_col("#entries [original]");
  if (!mode_uncompress)
    add_desc_col("#entries [multiple occurences compressed]");
  flush_desc(fp);

  if (mode_uncompress)
    print_data(fp,"%d,%d",total_freq,num_entry_pre);
  else
    print_data(fp,"%d,%d,%d",total_freq,num_entry_pre,num_entry);

  add_desc_id("PARAMSIZE");
  add_desc_col("Num. of classes");
  add_desc_col("Num. of class params");
  add_desc_col("Num. of params for discrete attrs");
  add_desc_col("Num. of params for continuous attrs");
  add_desc_col("Total num. of params");
  flush_desc(fp);

  if (mode_search_num_class) {
    for (k = kmin; k <= kmax; k += kstep) {
      num_param_d = 0;
      for (v=0; v<tuple_size_d; v++)
        num_param_d += k * (num_word[v] - 1);

      num_param_c = 0;
      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        num_param_c += k * (n + n * (n - 1) / 2);
      }

      total_num_param = (k - 1) + num_param_d + num_param_c;

      print_data(fp,"%d,%d,%d,%d,%d",
                 k,(k-1),num_param_d,num_param_c,total_num_param);
    }
  }
  else {
    k = num_class;

    num_param_d = 0;
    for (v=0; v<tuple_size_d; v++)
      num_param_d += k * (num_word[v] - 1);

    num_param_c = 0;
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];
      num_param_c += k * (n + n * (n - 1) / 2);
    }

    total_num_param = (k - 1) + num_param_d + num_param_c;

    print_data(fp,"%d,%d,%d,%d,%d",
               k,(k-1),num_param_d,num_param_c,total_num_param);
  }

  if (mode_nominal && total_num_miss_w > 0) {
    add_desc_id("NUM-MISSING-CAT");
    add_desc_col("Attr. ID");
    add_desc_col("Num. of missing values");
    add_desc_col("Ratio of missing values");
    flush_desc(fp);

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;

      v_d = attr_info[v].local_id;
      print_data(fp,"%d,%d,%.Ng",
                 v,num_miss_w[v_d],((FLOAT_T)num_miss_w[v_d])/total_freq);
    }
  }

  if (mode_gauss && total_num_miss_g > 0) {
    add_desc_id("NUM-MISSING-GAUSS");
    add_desc_col("Attr. ID");
    add_desc_col("Num. of missing values");
    add_desc_col("Ratio of missing values");
    flush_desc(fp);

    v_c = 0;
    for (v=0; v<tuple_size; v++) {
      if (is_attr_cont(v)) {
        print_data(fp,"%d,%d,%.Ng",
                   v,num_miss_g[v_c],((FLOAT_T)num_miss_g[v_c])/total_freq);
        v_c++;
      }
    }
  }

  fflush(fp);
}

void print_hparam_upd(FILE *fp)
{
  int k,v,v_d,i,j;
  int u,n,a,a0,a1;

  add_desc_id("HPARAM-CLASS");
  add_desc_col("Class k");
  add_desc_col("Class pseudo count delta(k)");
  flush_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,hparam_c_upd[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("HPARAM-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Class k");
    add_desc_col("Attr. value x%d",v);
    add_desc_col("Attr. pseudo count delta(x%d|k)",v);
    flush_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
        print_data(fp,"%d,%d,%s,%.Ng",
                   v,k,word_list[v_d][j],hparam_w_upd[k][v_d][j]);
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      add_desc_id("HPARAM-GAUSS-MEAN%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Attr. ID");
      add_desc_col("Mean");
      flush_desc(fp);

      for (k=0; k<num_class; k++)
        for (a=0; a<n; a++)
          print_data(fp,"%d,%d,%d,%.Ng",
                     u,k,gauss_attr[u][a],hparam_u_upd[k][u][a]);

      add_desc_id("HPARAM-GAUSS-COVAR%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Attr. ID");
      add_desc_col("Attr. ID");
      add_desc_col("(Co)Variance");
      flush_desc(fp);

      for (k=0; k<num_class; k++)
        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++)
            print_data(fp,"%d,%d,%d,%d,%.Ng",
                       u,k,gauss_attr[u][a0],gauss_attr[u][a1],
                       hparam_s_upd[k][u][a0][a1]);
    }

    add_desc_id("HPARAM-GAUSS-TAU");
    add_desc_col("Class k");
    add_desc_col("Tau");
    flush_desc(fp);

    for (k=0; k<num_class; k++)
      print_data(fp,"%d,%.Ng",k,(tau + expect_c[k]));

    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      add_desc_id("HPARAM-GAUSS-ALPHA%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Alpha");
      flush_desc(fp);

      for (k=0; k<num_class; k++)
        print_data(fp,"%d,%d,%.Ng",u,k,(get_alpha_w(n) + expect_c[k]));
    }
  }

  fflush(fp);
}

void print_pi(FILE *fp)
{
  int k,v,v_d,i,j;

  add_desc_id("PI-CLASS");
  add_desc_col("Class k");
  add_desc_col("Class pi(k)");
  flush_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,pi_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("PI-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Class k");
    add_desc_col("Attr. value x%d",v);
    add_desc_col("Attr. pi(x%d|k)",v);
    flush_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
        print_data(fp,"%d,%d,%s,%.Ng",v,k,word_list[v_d][j],pi_w[k][v_d][j]);
      }
    }
  }

  fflush(fp);
}

void print_complete_prob(FILE *fp)
{
  int k,v,e;

  add_desc_id("COMP");
  add_desc_col("Class k");
  add_desc_col("Case ID");
  if (!mode_hide_attr) add_desc_col("Attribute values x");
  add_desc_col("Complete data prob. P(k,x)");
  flush_desc(fp);

  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      add_data_int(k);
      add_data_str(get_case_id(e));
      if (!mode_hide_attr)
        for (v=0; v<tuple_size; v++)
          if (!is_attr_aux(v))
            add_data_str(freq_entry_val(e,v));
      add_data_float(prob_comp[k][e]);
      flush_data(fp);
    }
  }
  fflush(fp);
}

void print_incomplete_prob(FILE *fp)
{
  int v,e;

  add_desc_id("INCOMP");
  add_desc_col("Case ID");
  if (!mode_hide_attr) add_desc_col("Attribute values x");
  add_desc_col("Incomplete data prob. P(x)");
  flush_desc(fp);

  for (e=0; e<num_entry; e++) {
    add_data_str(get_case_id(e));
    if (!mode_hide_attr)
      for (v=0; v<tuple_size; v++)
        if (!is_attr_aux(v))
          add_data_str(freq_entry_val(e,v));
    add_data_float(prob_incomp[e]);
    flush_data(fp);
  }

  fflush(fp);
}

void print_conditional_prob(FILE *fp)
{
  int k,e,v;

  add_desc_id("COND");
  add_desc_col("Case ID");
  if (!mode_hide_attr) add_desc_col("Attribute values x");
  add_desc_col("Class k");
  add_desc_col("Conditional prob. P(k|x)");
  flush_desc(fp);

  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      add_data_str(get_case_id(e));
      if (!mode_hide_attr)
        for (v=0; v<tuple_size; v++)
          if (!is_attr_aux(v))
            add_data_str(freq_entry_val(e,v));
      add_data_int(k);
      add_data_float(prob_cond[k][e]);
      flush_data(fp);
    }
  }

  fflush(fp);
}

void print_expectation(FILE *fp)
{
  int k,v,v_d,i,j;

  add_desc_id("EXPECT-CLASS");
  add_desc_col("Class k");
  add_desc_col("Expected occurrences E[k]");
  flush_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,expect_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("EXPECT-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Class k");
    add_desc_col("Attr. value x%d",v);
    add_desc_col("Expected occurrences E[x%d|k]",v);
    flush_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
        print_data(fp,"%d,%d,%s,%.Ng",
                   v,k,word_list[v_d][j],expect_w[k][v_d][j]);
      }
    }
  }

  fflush(fp);
}

void print_decrease(FILE *fp, int r, int m,
                    FLOAT_T pre_free_energy, FLOAT_T post_free_energy,
                    FLOAT_T itemp0)
{
  add_desc_id("ERR-MSG-DECREASE");
  add_desc_comm("Variational free energy decreased!");
  flush_desc(fp);

  print_data(fp,"NumberOfClasses,%d",num_class);
  print_data(fp,"Restart,%d",r);
  print_data(fp,"NumberOfUpdates,%d",m);
  print_data(fp,"FreeEnergyBeforeUpdate,%.Ng",pre_free_energy);
  print_data(fp,"FreeEnergyAfterUpdate,%.Ng",post_free_energy);
  if (mode_daem) {
    print_data(fp,"InverseTemperature,%.Ng",itemp0);
  }
}

void print_decrease_sm(FILE *fp, int t, int m,
                       FLOAT_T pre_free_energy, FLOAT_T post_free_energy,
                       FLOAT_T itemp0)
{
  add_desc_id("ERR-MSG-DECREASE");
  add_desc_comm("Variational free energy decreased!");
  flush_desc(fp);

  print_data(fp,"NumberOfClasses,%d",num_class);
  print_data(fp,"SplitMergeOperation,%d",t);
  print_data(fp,"NumberOfUpdates,%d",m);
  print_data(fp,"FreeEnergyBeforeUpdate,%.Ng",pre_free_energy);
  print_data(fp,"FreeEnergyAfterUpdate,%.Ng",post_free_energy);
  if (mode_daem) {
    print_data(fp,"InverseTemperature,%.Ng",itemp0);
  }
}

void print_free_energy(FILE *fp, int k, int num_iterate)
{
  int m;

  if (mode_daem) {
    add_desc_id("FREE-ENERGY-K%d",k);
    add_desc_col("Num. of classes");
    add_desc_col("Num. of updates");
    add_desc_col("Variational free energy");
    add_desc_col("Inverse temperature");
    flush_desc(fp);

    for (m=0; m<num_iterate; m++) {
      print_data(fp,"%d,%d,%.Ng,%.Ng",k,m,free_energy[m],log_itemp[m]);
    }
  }
  else {
    add_desc_id("FREE-ENERGY-K%d",k);
    add_desc_col("Num. of classes");
    add_desc_col("Num. of updates");
    add_desc_col("Variational free energy");
    flush_desc(fp);

    for (m=0; m<num_iterate; m++) {
      print_data(fp,"%d,%d,%.Ng",k,m,free_energy[m]);
    }
  }

  fflush(fp);
}

void print_ic(FILE *fp)
{
  int k,u;

  add_desc_id("SCORE");
  add_desc_col("Num. of classes");
  add_desc_col("Variational free energy");
  if (mode_search_num_class)
    add_desc_comm("Optimal num. of classes = %d",num_class_opt);
  else
    add_desc_comm("Optimal num. of classes = %d",num_class);
  flush_desc(fp);

  if (mode_search_num_class) {
    u = 0;
    for (k = kmin; k <= kmax; k += kstep) {
      print_data(fp,"%d,%.Ng",k,ic[u]);
      u++;
    }
  }
  else print_data(fp,"%d,%.Ng",num_class,ic[0]);

  fflush(fp);
}

void print_ans_class_list(FILE *fp)
{
  int k;

  add_desc_id("ANS-CLASS");
  add_desc_col("Answer class");
  add_desc_col("Num of occurrences");
  flush_desc(fp);

  for (k=0; k<num_ans_class; k++)
    print_data(fp,"%s,%d",ans_class_list[k],cluster_case_freq_ans[k]);
}

void print_smooth(FILE *fp, int k)
{
  int v,v_d;

  add_desc_id("AUTO-SMOOTH-CLASS-K%d",k);
  add_desc_col("Num. of classes");
  add_desc_col("Class smoothing constant");
  flush_desc(fp);

  print_data(fp,"%d,%.Ng",k,smooth_c);

  add_desc_id("AUTO-SMOOTH-CAT-K%d",k);
  add_desc_col("Attr. ID");
  add_desc_col("Num. of classes");
  add_desc_col("Attr. smoothing constant");
  flush_desc(fp);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;
    v_d = attr_info[v].local_id;
    print_data(fp,"%d,%d,%.Ng",v,k,smooth_w[v_d]);
  }
}

void print_param(FILE *fp)
{
  int k,v,v_d,i,j;

  add_desc_id("PARAM-CLASS");
  add_desc_col("Class k");
  add_desc_col("Class parameter P(k)");
  flush_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,param_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("PARAM-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Class k");
    add_desc_col("Attr. value x%d",v);
    add_desc_col("Attr. parameter P(x%d|k)",v);
    flush_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
        print_data(fp,"%d,%d,%s,%.Ng",v,k,word_list[v_d][j],param_w[k][v_d][j]);
      }
    }
  }

  fflush(fp);
}

void print_case_member_prob(FILE *fp)
{
  int k,v,e;

  add_desc_id("MEMBER-CASE");
  add_desc_col("Case ID");
  if (mode_eval) add_desc_col("Answer class");
  if (!mode_hide_attr) add_desc_col("Attribute values x");
  add_desc_col("Class k");
  add_desc_col("Class prob. P(k)");
  add_desc_col("Component prob. P(x|k)");
  add_desc_col("Marginal prob. P(x)");
  add_desc_col("Membership prob. P(k|x)");
  flush_desc(fp);

  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      add_data_str(get_case_id(e));
      if (mode_eval) add_data_str(get_ans_class(e));
      if (!mode_hide_attr)
        for (v=0; v<tuple_size; v++)
          if (!is_attr_aux(v))
            add_data_str(freq_entry_val(e,v));
      add_data_int(k);
      add_data_float(param_c[k]);
      add_data_float(prob_compo_case[k][e]);
      add_data_float(prob_case[e]);
      add_data_float(prob_mem_case[k][e]);
      flush_data(fp);
    }
  }

  fflush(fp);
}

void print_attr_member_prob(FILE *fp)
{
  int k,v,v_d,i,j,u,a,e,n;

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;
    v_d = attr_info[v].local_id;

    add_desc_id("MEMBER-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Attr. value x%d",v);
    add_desc_col("Class k");
    add_desc_col("Class prob. P(k)");
    add_desc_col("Component prob. P(x%d|k)",v);
    add_desc_col("Marginal prob. P(x%d)",v);
    add_desc_col("Membership prob. P(k|x%d)",v);
    flush_desc(fp);

    for (i=0; i<num_word[v_d]; i++) {
      j = print_order[v_d][i];
      for (k=0; k<num_class; k++)
        print_data(fp,"%d,%s,%d,%.Ng,%.Ng,%.Ng,%.Ng",
                   v,word_list[v_d][j],k,
                   param_c[k],prob_compo_w[k][v_d][j],
                   prob_w[v_d][j],prob_mem_w[k][v_d][j]);
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      add_desc_id("MEMBER-GAUSS%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Case ID");
      if (!mode_hide_attr)
        for (a=0; a<n; a++)
          add_desc_col("Attr. x%d",gauss_attr[u][a]);
      add_desc_col("Class k");
      add_desc_col("Class prob. P(k)");
      add_desc_col("Component prob. P(Gauss%d|k)",u);
      add_desc_col("Marginal prob. P(Gauss%d)",u);
      add_desc_col("Membership prob. P(k|Gauss%d)",u);
      flush_desc(fp);

      for (e=0; e<num_entry; e++) {
        for (k=0; k<num_class; k++) {
          add_data_int(u);
          add_data_str(get_case_id(e));
          if (!mode_hide_attr)
            for (a=0; a<n; a++)
              add_data_str(freq_entry_val(e,gauss_attr[u][a]));
          add_data_int(k);
          add_data_float(param_c[k]);
          add_data_float(prob_compo_g[k][u][e]);
          add_data_float(prob_g[u][e]);
          add_data_float(prob_mem_g[k][u][e]);
          flush_data(fp);
        }
      }
    }
  }

  fflush(fp);
}

void print_predict_prob(FILE *fp)
{
  int k,u,n,v,v_d,i,j,a,a0,a1;
  FLOAT_T f,tau0;
  MATRIX m,**h_s;
  VECTOR **h_u;

  add_desc_id("PREDICT-CLASS");
  add_desc_col("Class k");
  add_desc_col("Class predictive prob. P(k)");
  flush_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%.Ng",k,param_c[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("PREDICT-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Class k");
    add_desc_col("Attr. value x%d",v);
    add_desc_col("Attr. predictive prob. P(x%d|k)",v);
    flush_desc(fp);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v_d]; i++) {
        j = print_order[v_d][i];
        print_data(fp,"%d,%d,%s,%.Ng",v,k,word_list[v_d][j],param_w[k][v_d][j]);
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      if (mode_em) {
        h_u = hparam_u_upd;
        h_s = hparam_s_upd;
      }
      else {
        h_u = hparam_u;
        h_s = hparam_s;
      }

      add_desc_id("PREDICT-GAUSS-MEAN%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Attr. ID");
      add_desc_col("Mean");
      flush_desc(fp);

      for (k=0; k<num_class; k++)
        for (a=0; a<n; a++)
          print_data(fp,"%d,%d,%d,%.Ng",u,k,gauss_attr[u][a],h_u[k][u][a]);

      add_desc_id("PREDICT-GAUSS-COVAR%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Attr. ID");
      add_desc_col("Attr. ID");
      add_desc_col("(Co)Variance");
      flush_desc(fp);

      for (k=0; k<num_class; k++) {
        if (mode_em) {
          tau0 = tau + expect_c[k];
          f = get_alpha_w(n) + expect_c[k] + 1 - n;
        }
        else {
          tau0 = hparam_tau[k];
          f = hparam_alpha_w[k][u] + 1 - n;
        }

        m = dup_matrix(h_s[k][u]);
        mul_const_matrix(m,((tau0 + 1) / (tau0 * f)));

        for (a0=0; a0<n; a0++)
          for (a1=a0; a1<n; a1++)
            print_data(fp,"%d,%d,%d,%d,%.Ng",
                       u,k,gauss_attr[u][a0],gauss_attr[u][a1],
                       m[a0][a1]);

        release_matrix(m);
      }

      add_desc_id("PREDICT-GAUSS-DOF%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Degrees of freedom");
      flush_desc(fp);

      for (k=0; k<num_class; k++) {
        if (mode_em)
          f = get_alpha_w(n) + expect_c[k] + 1 - n;
        else
          f = hparam_alpha_w[k][u] + 1 - n;

        print_data(fp,"%d,%d,%.Ng",u,k,f);
      }
    }
  }
}

void print_cluster(FILE *fp)
{
  int k,t,i,j,v,v_d,e,u,a,n;

  switch (cl_type) {

  case CL_MP:
    add_desc_id("CLUSTER-CASE-BY-CASE");
    add_desc_col("Case ID");
    if (mode_eval) add_desc_col("Answer class");
    add_desc_col("Cluster k");
    if (!mode_hide_attr) add_desc_col("Attribute values x");
    add_desc_col("Membership prob. P(k|x)");
    flush_desc(fp);

    for (e=0; e<num_entry; e++) {
      k = cluster_case_pred[e];
      add_data_str(get_case_id(e));
      if (mode_eval) add_data_str(get_ans_class(e));
      add_data_int(k);
      if (!mode_hide_attr)
        for (v=0; v<tuple_size; v++)
          if (!is_attr_aux(v))
            add_data_str(freq_entry_val(e,v));
      add_data_float(prob_mem_case[k][e]);
      flush_data(fp);
    }

    add_desc_id("CLUSTER-CASE-BY-CLUSTER");
    add_desc_col("Cluster k");
    if (mode_eval) add_desc_col("Answer class");
    add_desc_col("Case ID");
    if (!mode_hide_attr) add_desc_col("Attribute values x");
    add_desc_col("Membership prob. P(k|x)");
    flush_desc(fp);

    for (k=0; k<num_class; k++) {
      for (t=0; t<cluster_case_freq[k]; t++) {
        e = cluster_case[k][t];

        add_data_int(k);
        if (mode_eval) add_data_str(get_ans_class(e));
        add_data_str(get_case_id(e));
        if (!mode_hide_attr)
          for (v=0; v<tuple_size; v++)
            if (!is_attr_aux(v))
              add_data_str(freq_entry_val(e,v));
        add_data_float(prob_mem_case[k][e]);
        flush_data(fp);
      }
    }
    break;

  default:
    die("Internal error: Unsupported clustering method\n");
  }

  if (mode_show_attr_cluster && mode_nominal) {
    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;

      v_d = attr_info[v].local_id;

      switch (cl_type) {

      case CL_MP:
        add_desc_id("CLUSTER-CAT%d-BY-ATT",v);
        add_desc_col("Attr. ID");
        add_desc_col("Attr. x%d",v);
        add_desc_col("Cluster k");
        add_desc_col("Membership prob. P(k|x%d)",v);
        flush_desc(fp);

        for (i=0; i<num_word[v_d]; i++) {
          j = print_order[v_d][i];
          k = cluster_w_pred[v_d][j];
          print_data(fp,"%d,%s,%d,%.Ng",
                     v,word_list[v_d][j],k,prob_mem_w[k][v_d][j]);
        }

        add_desc_id("CLUSTER-CAT%d-BY-CLUSTER",v);
        add_desc_col("Attr. ID");
        add_desc_col("Cluster k");
        add_desc_col("Attr. x%d",v);
        add_desc_col("Membership prob. P(k|x%d)",v);
        flush_desc(fp);

        for (k=0; k<num_class; k++) {
          for (t=0; t<cluster_w_freq[v_d][k]; t++) {
            i = cluster_w[v_d][k][t];
            print_data(fp,"%d,%d,%s,%.Ng",
                       v,k,word_list[v_d][i],prob_mem_w[k][v_d][i]);
          }
        }
        break;

      default:
        die("Internal error: Unsupported clustering method\n");
      }
    }
  }

  if (mode_show_attr_cluster && mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      switch (cl_type) {

      case CL_MP:
        add_desc_id("CLUSTER-GAUSS%d-BY-ATT",u);
        add_desc_col("Gauss ID");
        add_desc_col("Case ID");
        if (!mode_hide_attr)
          for (a=0; a<n; a++)
            add_desc_col("Attr. x%d",gauss_attr[u][a]);
        add_desc_col("Cluster k");
        add_desc_col("Membership prob. P(k|Gauss%d)",u);
        flush_desc(fp);

        for (e=0; e<num_entry; e++) {
          k = cluster_g_pred[u][e];
          add_data_int(u);
          add_data_str(get_case_id(e));
          if (!mode_hide_attr)
            for (a=0; a<n; a++)
              add_data_str(freq_entry_val(e,gauss_attr[u][a]));
          add_data_int(k);
          add_data_float(prob_mem_g[k][u][e]);
          flush_data(fp);
        }

        add_desc_id("CLUSTER-GAUSS%d-BY-CLUSTER",u);
        add_desc_col("Gauss ID");
        add_desc_col("Cluster k");
        add_desc_col("Case ID");
        if (!mode_hide_attr)
          for (a=0; a<n; a++)
            add_desc_col("Attr. x%d",gauss_attr[u][a]);
        add_desc_col("Membership prob. P(k|Gauss%d)",u);
        flush_desc(fp);

        for (k=0; k<num_class; k++) {
          for (t=0; t<cluster_g_freq[u][k]; t++) {
            e = cluster_g[u][k][t];
            add_data_int(u);
            add_data_int(k);
            add_data_str(get_case_id(e));
            if (!mode_hide_attr)
              for (a=0; a<n; a++)
                add_data_str(freq_entry_val(e,gauss_attr[u][a]));
            add_data_float(prob_mem_g[k][u][e]);
            flush_data(fp);
          }
        }
        break;

      default:
        die("Internal error: Unsupported clustering method\n");
      }
    }
  }
}

void print_cluster_freq(FILE *fp)
{
  int k,v,v_d,u;

  add_desc_id("CLUSTER-FREQ-CASE");
  add_desc_col("Cluster for cases");
  add_desc_col("Cluster size");
  flush_desc(fp);

  for (k=0; k<num_class; k++)
    print_data(fp,"%d,%d",k,cluster_case_freq[k]);

  for (v=0; v<tuple_size; v++) {
    if (!is_attr_disc(v)) continue;

    v_d = attr_info[v].local_id;

    add_desc_id("CLUSTER-FREQ-CAT%d",v);
    add_desc_col("Attr. ID");
    add_desc_col("Cluster for attr. %d",v);
    add_desc_col("Cluster size");
    flush_desc(fp);

    for (k=0; k<num_class; k++)
      print_data(fp,"%d,%d,%d",v,k,cluster_w_freq[v_d][k]);
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      add_desc_id("CLUSTER-FREQ-GAUSS%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Cluster for Gauss%d",u);
      add_desc_col("Cluster size");
      flush_desc(fp);

      for (k=0; k<num_class; k++)
        print_data(fp,"%d,%d,%d",u,k,cluster_g_freq[u][k]);
    }
  }

  fflush(fp);
}

void print_rank_ml(FILE *fp)
{
  int k,k0,i,j,v,v_d,e,u,a,n;
  int num_show_rank_case;
  int *num_show_rank_w;

  num_show_rank_case =
    (!mode_rank_th && num_rank > 0 && num_rank < num_entry) ?
    num_rank : num_entry;

  add_desc_id("RANK-CASE");
  add_desc_col("Class k");
  if (mode_eval) add_desc_col("Answer class");
  add_desc_col("Case ID");
  if (!mode_hide_attr) add_desc_col("Attribute values x");
  add_desc_col("Class prob. P(k)");
  add_desc_col("Membership prob. P(k|x)");
  add_desc_comm("in the decending order of the class probs. and"
                " the membership probs.");
  flush_desc(fp);

  for (k=0; k<num_class; k++) {
    k0 = rank_class[k];
    for (e=0; e<num_show_rank_case; e++) {
      j = rank_mem_case[k0][e];
      if (mode_rank_th && prob_mem_case[k0][j] < rank_th) break;
      if (mode_rank_hide && prob_mem_case[k0][j] < 1.0/num_class) break;

      add_data_int(k0);
      if (mode_eval) add_data_str(get_ans_class(j));
      add_data_str(get_case_id(j));
      if (!mode_hide_attr)
        for (v=0; v<tuple_size; v++)
          if (!is_attr_aux(v))
            add_data_str(freq_entry_val(j,v));
      add_data_float(param_c[k0]);
      add_data_float(prob_mem_case[k0][j]);
      flush_data(fp);
    }
  }

  if (mode_nominal) {
    num_show_rank_w = (int *)xmalloc(sizeof(int)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++)
      num_show_rank_w[v] =
        (!mode_rank_th && num_rank > 0 && num_rank < num_word[v]) ?
        num_rank : num_word[v];

    for (v=0; v<tuple_size; v++) {
      if (!is_attr_disc(v)) continue;

      v_d = attr_info[v].local_id;

      add_desc_id("RANK-CAT%d",v);
      add_desc_col("Attr. ID");
      add_desc_col("Class k");
      add_desc_col("Attr. value x%d",v);
      add_desc_col("Class prob. P(k)");
      add_desc_col("Membership prob. P(k|x%d)",v);
      add_desc_comm("in the decending order of the class probs. and"
                    " the membership probs.");
      flush_desc(fp);

      for (k=0; k<num_class; k++) {
        k0 = rank_class[k];
        for (i=0; i<num_show_rank_w[v_d]; i++) {
          j = rank_mem_w[k0][v_d][i];
          if (mode_rank_th && prob_mem_w[k0][v_d][j] < rank_th) break;
          if (mode_rank_hide && prob_mem_w[k0][v_d][j] < 1.0/num_class) break;

          add_data_int(v);
          add_data_int(k0);
          add_data_str(word_list[v_d][j]);
          add_data_float(param_c[k0]);
          add_data_float(prob_mem_w[k0][v_d][j]);
          flush_data(fp);
        }
      }
    }

    xfree(num_show_rank_w);
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      add_desc_id("RANK-GAUSS%d",u);
      add_desc_col("Gauss ID");
      add_desc_col("Class k");
      add_desc_col("Case ID");
      if (!mode_hide_attr)
        for (a=0; a<n; a++)
          add_desc_col("Attr. x%d",gauss_attr[u][a]);
      add_desc_col("Class prob. P(k)");
      add_desc_col("Membership prob. P(k|Gauss%d)",u);
      add_desc_comm("in the decending order of the class probs. and"
                    " the membership probs.");
      flush_desc(fp);

      for (k=0; k<num_class; k++) {
        k0 = rank_class[k];
        for (e=0; e<num_show_rank_case; e++) {
          j = rank_mem_g[k0][u][e];
          if (mode_rank_th && prob_mem_g[k0][u][j] < rank_th) break;
          if (mode_rank_hide && prob_mem_g[k0][u][j] < 1.0/num_class) break;

          add_data_int(u);
          add_data_int(k0);
          add_data_str(get_case_id(j));
          if (!mode_hide_attr)
            for (a=0; a<n; a++)
              add_data_str(freq_entry_val(j,gauss_attr[u][a]));
          add_data_float(param_c[k0]);
          add_data_float(prob_mem_g[k0][u][j]);
          flush_data(fp);
        }
      }
    }
  }

  fflush(fp);
}

void print_rank(FILE *fp)
{
  switch (rel_m) {
  case REL_MP:
    print_rank_ml(fp);
    break;
  default:
    fprintf(stderr,
            "Internal error: Unexpected relevance measure between class"
            " and word\n");
    exit(1);
  }
}

void print_time_vb(FILE *fp)
{
  add_desc_id("CPUTIME");
  add_desc_col("CPU time [sec]");
  flush_desc(fp);

  print_data(fp,"Reading cases,%.Ng",(etime_read - stime_read));
  if (mode_em) print_data(fp,"EM,%.Ng",(etime_em - stime_em));
  if (mode_memp)
    print_data(fp,"Computing membership probs,%.Ng",
               (etime_memp - stime_memp));
  if (mode_cluster)
    print_data(fp,"Computing clusters,%.Ng",(etime_cluster - stime_cluster));
  if (mode_eval) print_data(fp,"Evaluation,%.Ng",(etime_eval - stime_eval));
  if (mode_rank)
    print_data(fp,"Relevance analysis,%.Ng",(etime_rank - stime_rank));
  if (mode_dissim) print_data(fp,"Computing dissimilarities,%.Ng",(etime_dissim - stime_dissim));

  fflush(fp);

  add_desc_id("LAPTIME");
  add_desc_col("Elapsed time [sec]");
  flush_desc(fp);

  print_data(fp,"Reading cases,%.Ng",(etime_lap_read - stime_lap_read));
  if (mode_em) print_data(fp,"EM,%.Ng",(etime_lap_em - stime_lap_em));
  if (mode_memp)
    print_data(fp,"Computing membership probs,%.Ng",
               (etime_lap_memp - stime_lap_memp));
  if (mode_cluster)
    print_data(fp,"Computing clusters,%.Ng",
               (etime_lap_cluster - stime_lap_cluster));
  if (mode_eval)
    print_data(fp,"Evaluation,%.Ng",(etime_lap_eval - stime_lap_eval));
  if (mode_rank)
    print_data(fp,"Relevance analysis,%.Ng",(etime_lap_rank - stime_lap_rank));
  if (mode_dissim)
    print_data(fp,"Computing dissimilarities,%.Ng",
               (etime_lap_dissim - stime_lap_dissim));

  fflush(fp);
}

void print_free_energy_sm(FILE *fp, int k)
{
  int t,m;

  if (mode_daem) {
    add_desc_id("FREE-ENERGY-SMEM-K%d",k);
    add_desc_col("Num. of classes");
    add_desc_col("Num. of split-merge operations");
    add_desc_col("Num. of updates");
    add_desc_col("Variational free energy");
    add_desc_col("Inverse temperature");
    flush_desc(fp);

    for (t=0; t<(num_sm+1); t++)
      for (m=0; m<num_iterate_sm_full[t]; m++)
        print_data(fp,"%d,%d,%d,%.Ng,%.Ng",
                   k,t,m,free_energy_sm[t][m],log_itemp_sm[t][m]);
  }
  else {
    add_desc_id("FREE-ENERGY-SMEM-K%d",k);
    add_desc_col("Num. of classes");
    add_desc_col("Num. of split-merge operations");
    add_desc_col("Num. of updates");
    add_desc_col("Variational free energy");
    flush_desc(fp);

    for (t=0; t<(num_sm+1); t++)
      for (m=0; m<num_iterate_sm_full[t]; m++)
        print_data(fp,"%d,%d,%d,%.Ng",k,t,m,free_energy_sm[t][m]);
  }
}

void print_num_result_sm(FILE *fp, int k)
{
  int t;

  add_desc_id("SMEM-STAT-SM-K%d",k);
  add_desc_col("Num. of classes");
  add_desc_col("Num. of split-merge operations");
  add_desc_col("Num. of rejected candidates");
  add_desc_col("Num. of ignored candidates");
  flush_desc(fp);

  for (t=0; t<num_sm; t++)
    print_data(fp,"%d,%d,%d,%d",
               k,t,num_result_rejected[t],num_result_skipped[t]);
}

void print_num_iterate_sm(FILE *fp, int k)
{
  int t;

  add_desc_id("SMEM-STAT-EM-K%d",k);
  add_desc_col("Num. of classes");
  add_desc_col("Num. of split-merge operations");
  add_desc_col("Num. of updates in partial EM");
  add_desc_col("Num. of updates in full EM");
  flush_desc(fp);

  for (t=0; t<(num_sm+1); t++) {
    if (t==0)
      print_data(fp,"%d,%d,N/A,%d",k,t,num_iterate_sm_full[t]);
    else
      print_data(fp,"%d,%d,%d,%d",
                 k,t,num_iterate_sm_part[t],
                 (num_iterate_sm_full[t]-num_iterate_sm_part[t]));
  }
}

void print_dissim_case_header(FILE *fp)
{
  add_desc_id("DISSIM-CASE");
  if (!mode_eval) {
    add_desc_col("Case ID of x");
    add_desc_col("Case ID of x'");
    add_desc_col("KL-divergence D(x||x')");
    flush_desc(fp);
  }
  else {
    add_desc_col("Case ID of x");
    add_desc_col("Answer class of x");
    add_desc_col("Case ID of x'");
    add_desc_col("Answer class of x'");
    add_desc_col("KL-divergence D(x||x')");
    flush_desc(fp);
  }
}

void print_dissim_w_header(FILE *fp, int v)
{
  add_desc_id("DISSIM-CAT%d",v);
  add_desc_col("Attr. ID");
  add_desc_col("Attr. value x%d",v);
  add_desc_col("Attr. value x%d'",v);
  add_desc_col("KL-divergence D(x%d||x%d')",v,v);
  flush_desc(fp);
}

void print_dissim_g_header(FILE *fp, int u)
{
  int n = gauss_attr_size[u], a;

  add_desc_id("DISSIM-GAUSS%d",u);

  add_desc_col("Gauss ID");
  add_desc_col("Case ID of Gauss%d",u);
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      add_desc_col("Attr. x%d",gauss_attr[u][a]);
  add_desc_col("Case ID of Gauss%d'",u);
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      add_desc_col("Attr. x%d'",gauss_attr[u][a]);
  add_desc_col("KL-divergence D(Gauss%d||Gauss%d')",u,u);

  flush_desc(fp);
}

void print_dissim_case(FILE *fp, int e0, int e1, FLOAT_T d)
{
  if (!mode_eval) {
    add_data_str(get_case_id(e0));
    add_data_str(get_case_id(e1));
    add_data_float(d);
    flush_data(fp);
  }
  else {
    add_data_str(get_case_id(e0));
    add_data_str(get_ans_class(e0));
    add_data_str(get_case_id(e1));
    add_data_str(get_ans_class(e1));
    add_data_float(d);
    flush_data(fp);
  }
}

void print_dissim_w(FILE *fp, int v, int v_d, int i, int j, FLOAT_T d)
{
  print_data(fp,"%d,%s,%s,%.Ng",v,word_list[v_d][i],word_list[v_d][j],d);
}

void print_dissim_g(FILE *fp, int u, int e0, int e1, FLOAT_T d)
{
  int n = gauss_attr_size[u], a;

  add_data_int(u);
  add_data_str(get_case_id(e0));
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      add_data_str(freq_entry_val(e0,gauss_attr[u][a]));
  add_data_str(get_case_id(e1));
  if (!mode_hide_attr)
    for (a=0; a<n; a++)
      add_data_str(freq_entry_val(e1,gauss_attr[u][a]));
  add_data_float(d);
  flush_data(fp);
}

void print_eval(FILE *fp)
{
  int k1,k2;

  add_desc_id("EVAL-PURITY");
  add_desc_col("Evaluation measure");
  add_desc_col("Value");
  add_desc_comm("Weight in F-measure of Purity = %.Ng",weight_purity);
  flush_desc(fp);

  print_data(fp,"Purity,%.Ng",purity);
  print_data(fp,"Inverse Purity,%.Ng",purity_inv);
  print_data(fp,"F-Purity,%.Ng",purity_f);

  add_desc_id("EVAL-RAND");
  add_desc_col("Evaluation measure");
  add_desc_col("Value");
  add_desc_comm("Weight in F-measure of Wallance's measures = %.Ng",
                weight_wallance);
  flush_desc(fp);

  print_data(fp,"Rand index,%.Ng",rand_index);
  print_data(fp,"Wallance's measure I,%.Ng",wallance1);
  print_data(fp,"Wallance's measure II,%.Ng",wallance2);
  print_data(fp,"F-measure of Wallance's measures,%.Ng",wallance_f);
  print_data(fp,"Fowlkes and Mallows index,%.Ng",fowlkes_mallows);
  print_data(fp,"Adjusted Rand index,%.Ng",adj_rand_index);
  print_data(fp,"Jaccard index,%.Ng",jaccard);
  print_data(fp,"Mirkin metric,%.Ng",mirkin);

  add_desc_id("EVAL-NMI");
  add_desc_col("Evaluation measure");
  add_desc_col("Value");
  flush_desc(fp);

  print_data(fp,"Normalized mutual information,%.Ng",nmi_score);
  print_data(fp,"Variation of information,%.Ng",vi_score);
  print_data(fp,"Normalized variation of information,%.Ng",nvi_score);

  add_desc_id("CLUSTER-MATCH-BY-PREDICTED");
  add_desc_col("Cluster k [predicted]");
  add_desc_col("Answer class");
  add_desc_col("Num. of matched cases");
  flush_desc(fp);

  for (k1=0; k1<num_class; k1++)
    for (k2=0; k2<num_ans_class; k2++)
      print_data(fp,"%d,%s,%d",
                 k1,ans_class_list[k2],cluster_case_freq_joint[k1][k2]);

  add_desc_id("CLUSTER-MATCH-BY-ANSWER");
  add_desc_col("Answer class");
  add_desc_col("Cluster k [predicted]");
  add_desc_col("Num. of matched cases");
  flush_desc(fp);

  for (k2=0; k2<num_ans_class; k2++)
    for (k1=0; k1<num_class; k1++)
      print_data(fp,"%s,%d,%d",
                 ans_class_list[k2],k1,cluster_case_freq_joint[k1][k2]);
}
