/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#ifdef VNBC
#  include "vnbc_def.h"
#  include "vnbc_ext.h"
#endif

/*
 * Count the number of non-comment lines
 * - this is not safe since we should not scan the file relying on '#'.
 */
static int get_number_of_entries(void)
{
  FILE *fp_freq;
  int c, pre_c;
  int ignore;
  int num_entry;

  fp_freq = fopen(fname_freq,"rb");
  if (fp_freq == NULL) die("Couldn't open %s\n",fname_freq);

  num_entry = 0;
  ignore = 0;
  pre_c = '\n';
  while ((c = getc(fp_freq)) != EOF) {
    if (pre_c == '\n' && c == '#') ignore = 1;
    if (c == '\n') {
      if (!ignore) num_entry++;
      ignore = 0;
    }
    pre_c = c;
  }
  if (pre_c != '\n') num_entry++;

  fclose(fp_freq);

  return num_entry;
}

static void alloc_freq_list(void)
{
  int e,v,u,a;

  freq_list = (struct FREQ *)xmalloc(sizeof(struct FREQ)*num_entry_pre);

  for (e=0; e<num_entry_pre; e++) {
    if (mode_nominal) {
      freq_list[e].id = (int *)xmalloc(sizeof(int)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) {
        freq_list[e].id[v] = WORD_NOT_ASSIGNED;
      }
      freq_list[e].miss_w = (int *)xmalloc(sizeof(int)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++) freq_list[e].miss_w[v] = 0;
    }

    if (mode_gauss) {
      freq_list[e].gval = (VECTOR *)xmalloc(sizeof(VECTOR)*gauss_size);
      for (u=0; u<gauss_size; u++)
        freq_list[e].gval[u] = alloc_vector(gauss_attr_size[u]);

      freq_list[e].obs_g = (int **)xmalloc(sizeof(int *)*gauss_size);
      freq_list[e].miss_g = (int **)xmalloc(sizeof(int *)*gauss_size);

      for (u=0; u<gauss_size; u++) {
        freq_list[e].obs_g[u] = (int *)xmalloc(sizeof(int)*gauss_attr_size[u]);
        freq_list[e].miss_g[u] = (int *)xmalloc(sizeof(int)*gauss_attr_size[u]);
        for (a=0; a<gauss_attr_size[u]; a++) {
          freq_list[e].obs_g[u][a] = 1;
          freq_list[e].miss_g[u][a] = 0;
        }
      }

      freq_list[e].num_miss_g = (int *)xmalloc(sizeof(int)*gauss_size);
      for (u=0; u<gauss_size; u++)
        freq_list[e].num_miss_g[u] = 0;
    }
    else {
      freq_list[e].gval = NULL;
      freq_list[e].obs_g = NULL;
      freq_list[e].miss_g = NULL;
      freq_list[e].num_miss_g = NULL;
    }

    freq_list[e].freq = 0;
    freq_list[e].line_no = -1;
    freq_list[e].ans_id = -1;
    freq_list[e].num_miss_w = 0;
  }
}

static void register_freq(int line_c, char **word, FLOAT_T *g_value,
                          int *miss_w, int *miss_g_value, int freq,
                          char *attr_vec, char *case_id,
                          char *ans_class, char *buf)
{
  int v,v_c,v_d,i,e,a,u;
  int lookup_only;
  int id,ans;

#if 0
  print_msg("buf: %s",buf);
  print_msg("  case_id = %s\n",case_id);
  print_msg("  ans_class = %s\n",ans_class);
  for (v_d=0; v_d<tuple_size_d; v_d++)
    print_msg("  word[%d] = %s (%d)\n",v_d,word[v_d],miss_w[v_d]);
  for (v_c=0; v_c<tuple_size_c; v_c++)
    print_msg("  g_value[%d] = %.Nf (%d)\n",
              v_c,g_value[v_c],miss_g_value[v_c]);
  print_msg("  freq = %d\n",freq);
  print_msg("  attr_vec = %s\n",attr_vec);
#endif

  /* for the data including ID or continuous values, we do not surpress
   * the entries:
   */
  if (mode_uncompress || mode_id || mode_gauss) e = entry_id++;
  else e = register_hashtable(&ht_vec, attr_vec);

  if (mode_nominal) {
    lookup_only = (!mode_as_is) ? 1 : 0;
    for (v_d=0; v_d<tuple_size_d; v_d++) {
      if (miss_w[v_d]) {
        freq_list[e].id[v_d] = WORD_MISSING;  /* indicates `missing' */
        freq_list[e].miss_w[v_d] = 1;
        freq_list[e].num_miss_w += 1;
        num_miss_w[v_d] += freq;
      }
      else {
        if (lookup_only)
          i = lookup_hashtable(&ht[v_d], word[v_d]);
        else
          i = register_hashtable(&ht[v_d], word[v_d]);

        if (i < 0) die("Unknown attribute value -- %s\n",word[v_d]);
        freq_list[e].id[v_d] = i;
      }
    }
  }

  if (mode_gauss) {
    v_c = 0;
    for (v=0; v<tuple_size; v++) {
      if (is_attr_cont(v)) {
        u = attr_info[v].gauss_id;
        a = attr_info[v].local_id;

        if (miss_g_value[v_c]) {
          freq_list[e].obs_g[u][a] = 0;
          freq_list[e].miss_g[u][a] = 1;
          freq_list[e].num_miss_g[u] += 1;
          num_miss_g[v_c] += freq;
        }
        else freq_list[e].gval[u][a] = g_value[v_c];

        v_c++;
      }
    }
  }

  if (freq_list[e].line_no < 0) freq_list[e].line_no = line_c;

  if (mode_id) {
    id = lookup_hashtable(&ht_case_id, case_id);

    if (id >= 0)
      die_line(fname_freq, line_c, __LINE__, buf, "duplicating case ID", -1);
    else
      id = register_hashtable(&ht_case_id, case_id);

    freq_list[e].case_id = id;
  }

  if (mode_eval) {
    ans = register_hashtable(&ht_class, ans_class);

    if (!mode_uncompress &&
        !mode_id &&
        !mode_gauss &&
        freq_list[e].num_miss_w == 0 &&
        freq_list[e].ans_id >= 0 &&
        freq_list[e].ans_id != ans)
      die_line(fname_freq, line_c, __LINE__, buf,
               "different answer classes for the same object", -1);

    freq_list[e].ans_id = ans;
  }

  freq_list[e].freq += freq;
}

static void make_word_freq_list(void)
{
  int i,e,v;

  word_freq_list = (int **)xmalloc(sizeof(int *)*tuple_size_d);

  for (v=0; v<tuple_size_d; v++) {
    word_freq_list[v] = (int *)xmalloc(sizeof(int)*num_word[v]);

    for (i=0; i<num_word[v]; i++)
      word_freq_list[v][i] = 0;

    for (e=0; e<num_entry; e++) {
      if (!freq_list[e].miss_w[v])
        word_freq_list[v][freq_list[e].id[v]] += freq_list[e].freq;
    }
  }
}

/*
 * [CSV form]
 *   - does not understand the double quatation character
 *   - allows the space/tab characters
 *   - ignores the lines starting with '#'
 *
 * [tab-separated from]
 *   - does not understand the double quatation character
 *   - allows the space characters
 *   - regards the consecutive tab characteres as one tab character
 *   - ignores the lines starting with '#'
 */
void read_freq_body(void)
{
  FILE *fp_freq;
  char *buf = NULL;
  char **word = NULL;
  char *attr_vec = NULL;
  char *case_id = NULL;
  char *ans_class = NULL;
  int freq,line_c,v,v_c,v_d;
  char *b,*endptr,*q;
  char *buf_conv = NULL;
  char **word_conv = NULL;
#ifdef USE_ICONV
  char *bb;
  char **ww = NULL;
  size_t buf_len_in, buf_len_out;
  size_t *w_len_in = NULL, *w_len_out = NULL;
#ifdef ICONV_CONST
  const char **word_tmp = NULL;
  const char *buf_tmp = NULL;
#else
  char **word_tmp = NULL;
  char *buf_tmp = NULL;
#endif
#endif
  FLOAT_T *g_value = NULL;
  int *miss_w = NULL;
  int *miss_g_value = NULL;
  char **field_freq = NULL;
  int *pos_freq = NULL;
  int max_field,num_field,has_freq,attr_vec_len;

  fp_freq = fopen(fname_freq,"rb");
  if (fp_freq == NULL) die("Couldn't open %s\n", fname_freq);

  buf = alloc_str(MAX_LINE_LEN + 1);

  if (mode_nominal) {
    word = (char **)xmalloc(sizeof(char *)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      word[v] = alloc_str(MAX_WORD_LEN + 1);
    miss_w = (int *)xmalloc(sizeof(int)*tuple_size_d);
  }

  if (mode_gauss) {
    g_value = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*tuple_size_c);
    miss_g_value = (int *)xmalloc(sizeof(int)*tuple_size_c);
  }

  if (mode_conv_in)
    buf_conv = alloc_str(MAX_LINE_LEN + 1);

  if (mode_id) case_id = alloc_str(MAX_WORD_LEN + 1);
  if (mode_eval) ans_class = alloc_str(MAX_WORD_LEN + 1);

  if (mode_nominal) {
#ifdef USE_ICONV
    w_len_in = (size_t *)xmalloc(sizeof(size_t)*tuple_size_d);
    w_len_out = (size_t *)xmalloc(sizeof(size_t)*tuple_size_d);
    ww = (char **)xmalloc(sizeof(char *)*tuple_size_d);

#ifdef ICONV_CONST
    word_tmp = (const char **)xmalloc(sizeof(const char *)*tuple_size_d);
#else
    word_tmp = (char **)xmalloc(sizeof(char *)*tuple_size_d);
#endif
#endif

    word_conv = (char **)xmalloc(sizeof(char *)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++)
      word_conv[v] = alloc_str(MAX_WORD_LEN + 1);
  }

  max_field = tuple_size + 1;

  field_freq = alloc_field(max_field);
  pos_freq = (int *)xmalloc(sizeof(int)*max_field);

  line_c = 0;
  while (fgets(buf,MAX_LINE_LEN,fp_freq) != NULL) {
    line_c++;

    for (v=0; v<tuple_size_d; v++) miss_w[v] = 0;
    for (v=0; v<tuple_size_c; v++) miss_g_value[v] = 0;

#ifdef USE_ICONV
    /* converted into the internal code if necessary */
    if (mode_conv_in) {
      buf_tmp = buf;
      bb = buf_conv;
      buf_len_in  = strlen(buf)+1;
      buf_len_out = MAX_LINE_LEN+1;
      if (iconv(cd_in, &buf_tmp, &buf_len_in, &bb, &buf_len_out) == (size_t)-1)
        die("[line %d] Couldn't convert to the internal charcode (errno=%d)\n",
            line_c,errno);
    }
    else {
      buf_conv = buf;
    }
#else
    buf_conv = buf;
#endif

    if (*buf_conv == '#') continue;

    num_field = -1;
    if (format_in == FMT_CSV)
      num_field = split_comma(max_field,buf_conv,field_freq,pos_freq);
    else
      num_field = split_tab(max_field,buf_conv,field_freq,pos_freq);

    if (num_field < 0)
      DIE_LINE_NFIELD(fname_freq,line_c,buf,"Too many fields (columns)");
    if (num_field < max_field - 1)
      DIE_LINE_NFIELD(fname_freq,line_c,buf,"Too few fields (columns)");

    has_freq = (num_field == max_field) ? 1 : 0;

    if (mode_id) strncpy(case_id,field_freq[attr_id],MAX_WORD_LEN);
    if (mode_eval) strncpy(ans_class,field_freq[attr_eval],MAX_WORD_LEN);

    freq = 1;
    v_c = 0;
    v_d = 0;
    for (v=0; v<num_field; v++) {
      if (is_attr_aux(v)) continue;

      if (strlen(field_freq[v]) == 0)
        die_line(fname_freq, line_c, __LINE__, buf_conv, "empty value",
                 pos_freq[v]);

      if (v == (num_field - 1) && has_freq) {
        b = field_freq[v];
        while (*b == '0') b++; /* convert "00083" --> "83", for example */

        freq = strtol(b, &endptr, 10);
        if (*endptr != '\0')
          die_line(fname_freq, line_c, __LINE__, buf_conv,
                   "illegal format for an integer", pos_freq[v]);
        if (freq <= 0)
          die_line(fname_freq, line_c, __LINE__, buf_conv,
                   "the object counts should be positive", pos_freq[v]);
        continue;
      }

      if (is_attr_cont(v)) {  /* for continuous values */
        if (strcmp(field_freq[v],"?") == 0 ||
            strcmp(field_freq[v],"NA") == 0) {
          miss_g_value[v_c] = 1;
        }
        else {
          g_value[v_c] = STRTOD(field_freq[v], &endptr);
          if (*endptr != '\0')
            die_line(fname_freq, line_c, __LINE__, buf_conv,
                     "illegal format for a floating-point number",
                     pos_freq[v]);
        }
        v_c++;
      }
      else { /* for discrete values */
        strncpy(word[v_d],field_freq[v],MAX_WORD_LEN);
        if (strcmp(word[v_d],"?") == 0 || strcmp(word[v_d],"NA") == 0)
          miss_w[v_d] = 1;
        v_d++;
      }
    }

    if (mode_nominal) {
#ifdef USE_ICONV
      /* converted into the output code if necessary */
      if (mode_conv_out) {
        for (v=0; v<tuple_size_d; v++) {
          word_tmp[v] = word[v];
          ww[v] = word_conv[v];
          w_len_in[v] = strlen(word[v])+1;
          w_len_out[v] = MAX_WORD_LEN+1;
          if (iconv(cd_out,&word_tmp[v],&w_len_in[v],&ww[v],&w_len_out[v]) == (size_t)-1 ||
              iconv(cd_out,&word_tmp[v],&w_len_in[v],&ww[v],&w_len_out[v]) == (size_t)-1)
            die("[line %d] Couldn't convert to %s (errno=%d)\n",
                line_c,charcode_out,errno);
        }
      }
      else {
        for (v=0; v<tuple_size_d; v++) {
          strncpy(word_conv[v],word[v],MAX_WORD_LEN);
        }
      }
#else
      for (v=0; v<tuple_size_d; v++) {
        strncpy(word_conv[v],word[v],MAX_WORD_LEN);
      }
#endif

      /* Construct a string representation of the discrete attribute vector: */
      attr_vec_len = (tuple_size_d - 1);
      for (v=0; v<tuple_size_d; v++)
        attr_vec_len += strlen(word_conv[v]);

      attr_vec = alloc_str(attr_vec_len + 1);

      b = attr_vec;
      for (v=0; v<tuple_size_d; v++) {
        if (v > 0) {
          *b = (format_in == FMT_CSV) ? ',' : '\t';
          b++;
        }
        q = word_conv[v];
        while (*q != '\0') {
          *b = *q;
          b++;
          q++;
        }
      }
      attr_vec[attr_vec_len] = '\0';
    }

    register_freq(line_c,word_conv,g_value,miss_w,miss_g_value,
                  freq,attr_vec,case_id,ans_class,buf_conv);

    if (mode_nominal) xfree(attr_vec);
  }

  xfree(buf);

  release_field(field_freq,max_field);
  xfree(pos_freq);
  xfree(attr_vec);

  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++)
      xfree(word[v]);
    xfree(word);

    if (mode_conv_in) xfree(buf_conv);

    for (v=0; v<tuple_size_d; v++)
      xfree(word_conv[v]);
    xfree(word_conv);

#ifdef USE_ICONV
    xfree(w_len_in);
    xfree(w_len_out);
    xfree(ww);
#  ifndef ICONV_CONST
    xfree(word_tmp);
#  endif
#endif

    xfree(miss_w);
  }

  if (mode_gauss) {
    xfree(g_value);
    xfree(miss_g_value);
  }

  if (mode_id) xfree(case_id);
  if (mode_eval) xfree(ans_class);

  fclose(fp_freq);
}

void read_freq(void)
{
  int e,f,v,i,k;

  print_msg("Reading cases from %s...", fname_freq);
  fflush(fp_msg);

  num_entry_pre = get_number_of_entries();
  if (num_entry_pre < 1) die("the file is empty\n");

  alloc_freq_list();

  if (mode_id) init_hashtable(&ht_case_id);
  if (mode_eval) init_hashtable(&ht_class);

  if (mode_nominal) {
    freq_w_id = (int *)xmalloc(sizeof(int)*tuple_size_d);

    /* The frequency file has lower priority than the parameter file: */
    if (mode_as_is) {
      ht = (HASHTABLE *)xmalloc(sizeof(HASHTABLE)*tuple_size_d);

      for (v=0; v<tuple_size_d; v++)
        init_hashtable(&ht[v]);
    }

    if (!(mode_uncompress || mode_id || mode_gauss))
      init_hashtable(&ht_vec);
  }

  if (mode_nominal) {
    num_word = (int *)xmalloc(sizeof(int)*tuple_size_d);

    num_miss_w = (int *)xmalloc(sizeof(int)*tuple_size_d);
    for (v=0; v<tuple_size_d; v++)
      num_miss_w[v] = 0;
  }

  if (mode_gauss) {
    num_miss_g = (int *)xmalloc(sizeof(int)*tuple_size_c);
    for (v=0; v<tuple_size_c; v++)
      num_miss_g[v] = 0;
  }

  entry_id = 0;
  read_freq_body();

  /* Get the number of actually distinct attribute vectors: */
  if (mode_uncompress || mode_id || mode_gauss) {
    if (num_entry_pre != entry_id)
      die("Internal error: inconsistency in number of entries");
    num_entry = entry_id;
  }
  else num_entry = get_cardinality(&ht_vec);

  /* Get the total number of cases: */
  f = 0;
  for (e=0; e<num_entry; e++)
    f += freq_list[e].freq;
  total_freq = f;

  if (mode_nominal) {
    /* Get the total number of missing values */
    total_num_miss_w = 0;
    for (v=0; v<tuple_size_d; v++)
      total_num_miss_w += num_miss_w[v];

    word_list = (char ***)xmalloc(sizeof(char **)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++) {
      num_word[v] = get_cardinality(&ht[v]);
      word_list[v] = get_keys(&ht[v]);
    }
  }

  if (mode_gauss) {
    total_num_miss_g = 0;
    for (v=0; v<tuple_size_c; v++)
      total_num_miss_g += num_miss_g[v];
  }

  /* Get the number of case IDs (= number of cases) in the evaluation data: */
  if (mode_id) {
    num_case_id  = get_cardinality(&ht_case_id);
    case_id_list = get_keys(&ht_case_id);
  }

  /* Get the number of classes in the evaluation data: */
  if (mode_eval) {
    num_ans_class  = get_cardinality(&ht_class);
    ans_class_list = get_keys(&ht_class);

    cluster_case_freq_ans = (int *)xmalloc(sizeof(int)*num_ans_class);
    for (k=0; k<num_ans_class; k++)
      cluster_case_freq_ans[k] = 0;

    for (e=0; e<num_entry; e++)
      cluster_case_freq_ans[freq_list[e].ans_id]++;
  }

  if (mode_nominal) {
    print_order = (int **)xmalloc(sizeof(int *)*tuple_size_d);

    for (v=0; v<tuple_size_d; v++) {
      print_order[v] = (int *)xmalloc(sizeof(int)*num_word[v]);
      for (i=0; i<num_word[v]; i++)
        print_order[v][i] = i;

      if (mode_lex_order) {
        tmp_word_list = word_list[v];
        qsort(print_order[v],num_word[v],sizeof(int),
              (int(*)(const void *,const void *))compare_word);
      }
    }

    make_word_freq_list();
  }

  if (mode_em) compute_stats();

  if (mode_eval) print_ans_class_list(fp_log);

  if (mode_verbose) {
    print_attr_freq_list(fp_log);
    print_case_freq_list(fp_log);
  }

  print_msg("done\n");
}

int compare_word(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  char *word_x = tmp_word_list[ix];
  char *word_y = tmp_word_list[iy];

  return strncmp(word_x,word_y,MAX_WORD_LEN);
}
