/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

static FLOAT_T **tmp_jmerge;
static FLOAT_T *tmp_jsplit;

void vbsmem_iterate(int u)
{
  int t,h;
  FLOAT_T last_free_energy;
  FLOAT_T curr_jsplit, curr_jmerge;
  int sm_ok;
  int n,n0;
  int m1,m2;
  int k1,k2,k3,last_k1=-1,last_k2=-1;
  int num_tried_cand, num_tried_jsplit_cand=-1;
  int num_iterate,num_iterate_best=-1;
  int r,r_best=-1;
  unsigned int curr_seed=0;
  int *cand;

  num_candidate = num_class*(num_class - 1)*(num_class - 2)/2;
  n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  if (num_init > 1)
    n0 = (max_init_iterate > 0) ? max_init_iterate : n;
  else
    n0 = n;

  alloc_jmerge();
  alloc_jsplit();
  alloc_candidate();
  alloc_hparam_sm();
  alloc_free_energy_sm_size();
  alloc_free_energy_sm();
  if (mode_daem) alloc_log_itemp_sm_vb();
  alloc_num_iterate_sm();
  alloc_num_result_sm();
  alloc_prob_part();
  alloc_param();
  alloc_prob_mem();

  for (t=0; t<(num_sm+1); t++) {
    num_iterate_sm_full[t] = 0;
    num_iterate_sm_part[t] = 0;
  }

  for (t=0; t<num_sm; t++) {
    num_result_rejected[t] = 0;
    num_result_skipped[t] = 0;
  }

  for (r=0; r<num_init; r++) {
    if (!mode_set_seed) {
      curr_seed = (unsigned int)random_int();
      random_seed(curr_seed);
    }

    print_msg("  [%d] #iters ",r);

    initialize_hparam();

    num_iterate = vbsmem_iterate_full(0,0,n0,itemp_init);

    num_iterate_sm_full[0] = num_iterate;

    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_free_energy[u][r] = final_free_energy;

    if (r == 0 || final_free_energy > final_free_energy_best) {
      r_best = r;
      final_free_energy_best = final_free_energy;
      num_iterate_best = num_iterate;
      save_free_energy_best_sm(0,num_iterate_best);
      if (mode_daem) save_log_itemp_best_sm(0,num_iterate_best);
      save_hparam_best();
    }
  }
  log_r_best[u] = r_best;

  final_free_energy = final_free_energy_best;
  restore_hparam_best();
  restore_free_energy_best_sm(0,num_iterate_best);
  if (mode_daem) restore_log_itemp_best_sm(0,num_iterate_best);

  last_free_energy = final_free_energy;

  for (t=0; t<num_sm; t++) {
    save_hparam_sm();

    update_param_vbem();
    compute_case_member_prob();
    compute_jmerge();
    compute_jsplit();
    sort_candidate();

    sm_ok = 0;
    num_tried_cand = 0;
    for (h=0; (max_candidate==0 || num_tried_cand<max_candidate) && h<num_candidate; h++) {

      cand = sm_candidate[h];
      k1 = cand[0];
      k2 = cand[1];
      k3 = cand[2];
      curr_jmerge = jmerge[k1][k2];
      curr_jsplit = jsplit[k3];

      if (h==0 || last_k1 != k1 || last_k2 != k2) {
        num_tried_jsplit_cand = 0;
      }
      else {
        num_tried_jsplit_cand++;
      }

      if (max_jsplit>0 && num_tried_jsplit_cand >= max_jsplit) {
#ifdef PRINT_SKIPPED
        print_msg("    -- Skipped (the number of examined split candidates"
                  " reached the limit)\n");
#endif
        num_result_skipped[t]++;
        continue;
      }
      if ((min_param_c>0 &&
           param_c[k1] < min_param_c && param_c[k2] < min_param_c)
          || (min_param_c>0 && param_c[k3] < min_param_c)) {
#ifdef PRINT_SKIPPED
        print_msg("    -- Skipped (too small class probability)\n");
#endif
        num_result_skipped[t]++;
        continue;
      }

      print_msg("  [Op=%d]\n    Candidate %d (%d not skipped): "
                "Jmerge(%d,%d)=%.Nf / Jsplit(%d)=%.Nf\n",
                (t+1),h,num_tried_cand,k1,k2,curr_jmerge,k3,curr_jsplit);

      num_tried_cand++;
      last_k1 = k1;
      last_k2 = k2;

      print_msg("    #iters ");

      m1 = vbsmem_iterate_part((t+1),h,0,n);
      m2 = vbsmem_iterate_full((t+1),m1,n,1.0);

      if (final_free_energy > last_free_energy) {
        if ((final_free_energy - last_free_energy) > min_smem_improve) {
          if (!(min_new_param_c > 0) ||
              (param_c[k1] > min_new_param_c &&
               param_c[k2] > min_new_param_c &&
               param_c[k3] > min_new_param_c)) {

            print_msg("    -- Accepted [%d,%d]->[%d] & [%d]->[%d,%d] "
                      "(F current: %.Nf / last: %.Nf)\n",
                      k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
            sm_ok = 1;
          }
          else {
            print_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
                      "(F current: %.Nf / last: %.Nf)\n",
                      k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
            print_msg("       <Too small split/merged class probabilities: "
                      "%d - %.Nf, %d - %.Nf, %d - %.Nf>\n",
                      k1,param_c[k1],k2,param_c[k2],k3,param_c[k3]);
            sm_ok = 0;
          }
        }
        else {
          print_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
                    "(F current: %.Nf / last: %.Nf)\n",
                    k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
          print_msg("       <Too small improvement: %.Ne>\n",
                    (final_free_energy - last_free_energy));
          sm_ok = 0;
        }
      }
      else {
        print_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
                  "(F current: %.Nf / last: %.Nf)\n",
                  k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
        print_msg("       <No improvement>\n");
        sm_ok = 0;
      }

      if (sm_ok) {
        last_free_energy = final_free_energy;
        num_iterate_sm_part[t+1] = m1;
        num_iterate_sm_full[t+1] = m2;
        break;
      }
      else {
        num_result_rejected[t]++;
        restore_hparam_sm(); /* backtrack */
      }

    }

    if (sm_ok == 0) {
      if (h == num_candidate)
        print_msg("  [Op=%d]\n  <<All candidates are rejected>>\n",(t+1));
      else
        print_msg("  [Op=%d]\n  <<The number of rejected split-merge"
                  " candidates reached the limit>>\n",(t+1));

      break;
    }
  }

  print_num_iterate_sm(fp_log,num_class);
  print_free_energy_sm(fp_log,num_class);
  print_num_result_sm(fp_log,num_class);

  release_jmerge();
  release_jsplit();
  release_candidate();
  release_hparam_sm();
  release_free_energy_sm();
  if (mode_daem) release_log_itemp_sm_vb();
  release_free_energy_sm_size();
  release_num_iterate_sm();
  release_num_result_sm();
  release_prob_part();
}

int vbsmem_iterate_full(int t, int m, int n, FLOAT_T itemp_init0)
{
  int mm;

  itemp = (mode_daem) ? itemp_init0 : 1.0;

  compute_pi();
  if (mode_daem && itemp < 1.0)
    compute_complete_prob_daem();
  else
    compute_complete_prob();
  compute_incomplete_prob();

  free_energy_sm[t][m] =
    (mode_daem && itemp < 1.0) ?
    compute_free_energy_daem() : compute_free_energy();
  if (mode_daem) log_itemp_sm[t][m] = itemp;

  while (1) { /* Loop for inverse temperature */

    print_daem_message_sm(t,m,itemp);
    mm = 0;

    while (1) {
      print_dot_message(m);

      compute_conditional_prob();
      compute_expect();
      update_hparam();

      m++;
      mm++;

      compute_pi();
      if (mode_daem && itemp < 1.0)
        compute_complete_prob_daem();
      else
        compute_complete_prob();
      compute_incomplete_prob();
      if (m >= free_energy_sm_size[t]) expand_free_energy_sm_all(t);
      free_energy_sm[t][m] =
        (mode_daem && itemp < 1.0) ?
        compute_free_energy_daem() : compute_free_energy();
      if (mode_daem) log_itemp_sm[t][m] = itemp;

      if (is_decreased_sm(t,m,n)) break;

      if (mode_daem && is_converged_sm_daem(t,m,mm,n,itemp)) {
        break;
      }
      else if (!mode_daem && is_converged_sm(t,m,n)) {
        break;
      }
    }

    if (!mode_daem && m >= n) break;

    if (itemp < 1.0) itemp *= itemp_rate;
    else break;
    if (itemp > 1.0) itemp = 1.0;
  }

  final_free_energy = free_energy_sm[t][m];

  return m;
}

int is_decreased_sm(int t, int m, int n)
{
  /* This would not happen without a bug or a numeric error */
  if (mode_daem &&
      log_itemp_sm[t][m] - log_itemp_sm[t][m-1] < ACCEPT_SMALL_NUMBER &&
      free_energy_sm[t][m] < free_energy_sm[t][m-1] &&
      (free_energy_sm[t][m-1] - free_energy_sm[t][m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease_sm(fp_log,t,m,free_energy_sm[t][m-1],
                      free_energy_sm[t][m],log_itemp_sm[t][m]);
    die(conv_msg("\n(iter:%d) Variational free energy decreased! "
                 "(before: %.Nf (%.Nf) | after: %.Nf (%.Nf) | diff: %.Nf) "
                 "-- see the log file\n"),
        m,free_energy_sm[t][m-1],log_itemp_sm[t][m-1],
        free_energy_sm[t][m],log_itemp_sm[t][m],
        (free_energy_sm[t][m-1] - free_energy_sm[t][m]));
    return 1;
  }
  else if (!mode_daem &&
           free_energy_sm[t][m] < free_energy_sm[t][m-1] &&
           (free_energy_sm[t][m-1] - free_energy_sm[t][m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease_sm(fp_log,t,m,free_energy_sm[t][m-1],
                      free_energy_sm[t][m],0.0);
    die(conv_msg("\n(iter:%d) Variational free energy decreased! "
                 "(before: %.Nf | after: %.Nf | diff: %.Nf)"
                 " -- see the log file\n"),
        m,free_energy_sm[t][m-1],free_energy_sm[t][m],(free_energy_sm[t][m-1] - free_energy_sm[t][m]));
    return 1;
  }
  return 0;
}

int is_converged_sm(int t, int m, int n)
{
  if (m >= n) {
    print_msg(" (Stopped: %d iters)\tF=%.Nf\n",m,free_energy_sm[t][m]);
    return 1;
  }
  else if (free_energy_sm[t][m] - free_energy_sm[t][m-1] > 0 &&
           free_energy_sm[t][m] - free_energy_sm[t][m-1] < epsilon) {
    print_msg(" (Converged: %d iters)\tF=%.Nf\n",m,free_energy_sm[t][m]);
    return 1;
  }
  return 0;
}

/* m: total number of iterations,
 * mm: number of iterations at the current inverse temperature
 * itemp: should be =< 1.0
 */
int is_converged_sm_daem(int t, int m, int mm, int n, FLOAT_T itemp)
{
  if (itemp < 1.0) {
    if (mm >= n) {
      return 1;
    }
    else if (log_itemp_sm[t][m] - log_itemp_sm[t][m-1] < ACCEPT_SMALL_NUMBER &&
             free_energy_sm[t][m] - free_energy_sm[t][m-1] > 0 &&
             free_energy_sm[t][m] - free_energy_sm[t][m-1] < epsilon) {
      return 1;
    }
  }
  else {
    if (mm >= n) {
      print_msg(" (Stopped: %d iters)\tF=%.Nf\n",m,free_energy_sm[t][m]);
      return 1;
    }
    else if (log_itemp_sm[t][m] - log_itemp_sm[t][m-1] < ACCEPT_SMALL_NUMBER &&
             free_energy_sm[t][m] - free_energy_sm[t][m-1] > 0 &&
             free_energy_sm[t][m] - free_energy_sm[t][m-1] < epsilon) {
      print_msg(" (Converged: %d iters)\tF=%.Nf\n",m,free_energy_sm[t][m]);
      return 1;
    }
  }

  return 0;
}

int vbsmem_iterate_part(int t, int h, int m, int n)
{
  int *cand;

  cand = sm_candidate[h];

  compute_pi();
  compute_complete_prob();
  compute_incomplete_prob();

  initialize_hparam_part(cand);

  compute_pi();
  compute_complete_prob();
  compute_incomplete_prob();
  save_incomplete_prob_part(cand);

  free_energy_sm[t][m] = compute_free_energy();

  if (mode_daem) log_itemp_sm[t][m] = itemp;

  print_dot_message(m);

  while (1) {
    compute_conditional_prob_part(cand);
    compute_expect_part(cand);
    update_hparam_part(cand);

    m++;

    compute_pi_part(cand);
    compute_complete_prob_part(cand);
    compute_incomplete_prob_part(cand);

    free_energy_sm[t][m] = compute_free_energy();
    if (mode_daem) log_itemp_sm[t][m] = itemp;

    print_dot_message(m);

    if (free_energy_sm[t][m] < free_energy_sm[t][m-1]
        && free_energy_sm[t][m-1] - free_energy_sm[t][m] > ACCEPT_SMALL_NUMBER) {
      print_msg("\n(part) (iter:%d) Variational free energy decreased! "
                "(curr: %.Nf / last: %.Nf / diff: %.Nf)\n",
                m,free_energy_sm[t][m],free_energy_sm[t][m-1],
                (free_energy_sm[t][m-1] - free_energy_sm[t][m]));
      break;
    }

    if (m >= n) {
      print_msg("(S:%d)",m);
      break;
    }
    else {
      if (free_energy_sm[t][m] - free_energy_sm[t][m-1] < epsilon) {
        print_msg("(C:%d)",m);
        break;
      }
    }
  }

  /* No need to update final_free_energy, etc.  */

  return m;
}

void compute_jmerge(void)
{
  int k,k1,k2,e;
  FLOAT_T *norm;

  if (mode_norm) {
    norm = (FLOAT_T *)xmalloc(sizeof(FLOAT_T)*num_class);

#pragma omp parallel for private(e)
    for (k=0; k<num_class; k++) {
      norm[k] = 0.0;
      for (e=0; e<num_entry; e++) {
        norm[k] +=
          freq_list[e].freq * prob_mem_case[k][e] * prob_mem_case[k][e];
      }
    }

#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
        jmerge[k1][k2] = 0.0;
        for (e=0; e<num_entry; e++) {
          jmerge[k1][k2] +=
            freq_list[e].freq * prob_mem_case[k1][e] * prob_mem_case[k2][e];
        }
        jmerge[k1][k2] /= SQRT(norm[k1]*norm[k2]);
      }
    }

    xfree(norm);
  }
  else {
#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
        jmerge[k1][k2] = 0.0;
        for (e=0; e<num_entry; e++) {
          jmerge[k1][k2] +=
            freq_list[e].freq * prob_mem_case[k1][e] * prob_mem_case[k2][e];
        }
      }
    }

  }

}

void compute_jsplit(void)
{
  int k,e,v,u;
  FLOAT_T e_k,div,p_x,p_lim;

  p_lim = TINY_FLOAT * TINY_FLOAT;

  for (k=0; k<num_class; k++) {
    e_k = 0.0;
    for (e=0; e<num_entry; e++)
      e_k += prob_mem_case[k][e];
    if (!(e_k > 0)) e_k = p_lim;

    div = 0.0;
    for (e=0; e<num_entry; e++) {
      p_x = 0.0;
      if (mode_nominal) {
        for (v=0; v<tuple_size_d; v++)
          if (!freq_list[e].miss_w[v])
            p_x += LOG(param_w[k][v][freq_list[e].id[v]]);
      }
      if (mode_gauss) {
        for (u=0; u<gauss_size; u++)
          p_x += compute_predict_gauss(k,u,e);
      }
      if (!(prob_mem_case[k][e] > 0)) continue;
      if (!(p_x > 0)) p_x = LOG(p_lim);

      div += prob_mem_case[k][e] * (LOG(prob_mem_case[k][e]) - p_x);
    }

    jsplit[k] = div / e_k - LOG(e_k);
  }

}

static void compute_hparam_precisions_part(int *cand)
{
  int l,k,u;

  for (l=0; l<3; l++) {
    k = cand[l];
    for (u=0; u<gauss_size; u++) {
      hparam_sdet_upd[k][u] = INVERT(hparam_s_upd[k][u],hparam_p_upd[k][u]);
      if (err_inv)
        DIE_NUMERIC("Failed to invert the hyperparameters for covariances\n");
    }
  }

}

void initialize_hparam_part(int *cand)
{
  int i,k1,k2,k3,v,u,n,a,a0,a1;
  FLOAT_T r1,r2;
  FLOAT_T noise;
  FLOAT_T p1,p2,p3,q1,q2,q3,u1,u2,u3,s1,s2;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  q1 = hparam_c_upd[k1];
  q2 = hparam_c_upd[k2];
  q3 = hparam_c_upd[k3];

  hparam_c_upd[k1] = hparam_c[k1] + (q1 - hparam_c[k1]) + (q2 - hparam_c[k2]);
  hparam_c_upd[k2] = hparam_c[k2] + 0.5 * (q3 - hparam_c[k3]);
  hparam_c_upd[k3] = hparam_c[k3] + 0.5 * (q3 - hparam_c[k3]);

  r1 = expect_c[k1]/(expect_c[k1] + expect_c[k2]);
  r2 = expect_c[k2]/(expect_c[k1] + expect_c[k2]);

  /* To keep the repeatability, we do not use OpenMP here */
  if (mode_nominal) {
    for (v=0; v<tuple_size_d; v++) {
      for (i=0; i<num_word[v]; i++) {
        p1 = hparam_w_upd[k1][v][i] - hparam_w[k1][v][i];
        p2 = hparam_w_upd[k2][v][i] - hparam_w[k2][v][i];
        p3 = hparam_w_upd[k3][v][i] - hparam_w[k3][v][i];
        noise = noise_ratio * random_float() / num_word[v];
        hparam_w_upd[k1][v][i] = hparam_w[k1][v][i] + r1 * p1 + r2 * p2;
        hparam_w_upd[k2][v][i] = hparam_w[k2][v][i] + p3 * (1.0 - noise);
        hparam_w_upd[k3][v][i] = hparam_w[k3][v][i] + p3 * (1.0 + noise);
      }
    }
  }

  if (mode_gauss) {
    for (u=0; u<gauss_size; u++) {
      n = gauss_attr_size[u];

      for (a=0; a<n; a++) {
        u1 = hparam_u_upd[k1][u][a];
        u2 = hparam_u_upd[k2][u][a];
        u3 = hparam_u_upd[k3][u][a];
        noise =
          noise_ratio_gauss * SQRT(sample_s[u][a][a]) * (0.5 - random_float());
        hparam_u_upd[k1][u][a] = r1 * u1 + r2 * u2;
        hparam_u_upd[k2][u][a] = u3 + noise;
        hparam_u_upd[k3][u][a] = u3 - noise;
      }
      for (a0=0; a0<n; a0++) {
        for (a1=a0; a1<n; a1++) {
          s1 = hparam_s_upd[k1][u][a0][a1] - hparam_s[k1][u][a0][a1];
          s2 = hparam_s_upd[k2][u][a0][a1] - hparam_s[k2][u][a0][a1];
          hparam_s_upd[k1][u][a0][a1] =
            hparam_s[k1][u][a0][a1] + r1 * s1 + r2 * s2;
          hparam_s_upd[k2][u][a0][a1] = hparam_s_upd[k3][u][a0][a1];
          hparam_s_upd[k3][u][a0][a1] = hparam_s_upd[k3][u][a0][a1];
        }
      }
      copy_upper_to_lower(hparam_s_upd[k1][u]);
      copy_upper_to_lower(hparam_s_upd[k2][u]);
      copy_upper_to_lower(hparam_s_upd[k3][u]);
    }

    compute_hparam_precisions_part(cand);

    expect_c[k1] = expect_c[k1] + expect_c[k2];
    expect_c[k2] = 0.5 * expect_c[k3];
    expect_c[k3] = 0.5 * expect_c[k3];
  }
}

static int compare_candidate(const int **x, const int **y)
{
  int k1_x = (*x)[0];
  int k2_x = (*x)[1];
  int k3_x = (*x)[2];
  int k1_y = (*y)[0];
  int k2_y = (*y)[1];
  int k3_y = (*y)[2];

  FLOAT_T jm_x = tmp_jmerge[k1_x][k2_x];
  FLOAT_T jm_y = tmp_jmerge[k1_y][k2_y];
  FLOAT_T js_x = tmp_jsplit[k3_x];
  FLOAT_T js_y = tmp_jsplit[k3_y];

  if (k1_x == k1_y && k2_x == k2_y) {
    if (js_x > js_y) return -1;
    if (js_x < js_y) return 1;
    return 0;
  }
  else {
    if (jm_x > jm_y) return -1;
    if (jm_x < jm_y) return 1;
    return 0;
  }
}

void sort_candidate(void)
{
  int k1,k2,k3;
  int h;

  h = 0;
  for (k1=0; k1<num_class; k1++) {
    for (k2=(k1+1); k2<num_class; k2++) {
      for (k3=0; k3<num_class; k3++) {
        if (k3 == k1 || k3 == k2) continue;
        sm_candidate[h][0] = k1;
        sm_candidate[h][1] = k2;
        sm_candidate[h][2] = k3;
        h++;
      }
    }
  }

  tmp_jmerge = jmerge;
  tmp_jsplit = jsplit;

  qsort(sm_candidate,num_candidate,sizeof(int *),
        (int(*)(const void *, const void *))compare_candidate);
}

void compute_pi_part(int *cand)
{
  int l,k,v,i;

  FLOAT_T sum;
  FLOAT_T psi;

  sum = 0.0;
  for (k=0; k<num_class; k++) {
    sum += hparam_c_upd[k];
  }
  psi = DIGAMMA(sum);

  for (l=0; l<3; l++) {
    k = cand[l];
    pi_c[k] = DIGAMMA(hparam_c_upd[k]) - psi;
  }

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];

      for (v=0; v<tuple_size_d; v++) {
        sum = 0.0;
        for (i=0; i<num_word[v]; i++) {
          sum += hparam_w_upd[k][v][i];
        }
        psi = DIGAMMA(sum);

        for (i=0; i<num_word[v]; i++) {
          pi_w[k][v][i] = DIGAMMA(hparam_w_upd[k][v][i]) - psi;
        }
      }
    }
  }
}

void compute_complete_prob_part(int *cand)
{
  int l,k,v,e,u,n,j;
  int n_miss,n_obs;
  int *obs;
  FLOAT_T lp,e_c_k,alpha_w0,sum,xval;
  VECTOR h_u_upd,h_u_upd_obs,x;
  MATRIX h_s_upd,h_p_upd,h_s_upd_obs,h_s_upd_obs_inv;
  FLOAT_T h_sdet_upd,h_sdet_upd_obs;

  for (l=0; l<3; l++) {
    k = cand[l];

    for (e=0; e<num_entry; e++)
      prob_comp[k][e] = pi_c[k];

    if (mode_nominal) {
      for (e=0; e<num_entry; e++) {
        if (total_num_miss_w > 0 && freq_list[e].num_miss_w > 0) {
          for (v=0; v<tuple_size_d; v++) {
            if (!freq_list[e].miss_w[v])
              prob_comp[k][e] += pi_w[k][v][freq_list[e].id[v]];
          }
        }
        else {
          for (v=0; v<tuple_size_d; v++) {
            prob_comp[k][e] += pi_w[k][v][freq_list[e].id[v]];
          }
        }
      }
    }

    if (mode_gauss) {
      e_c_k = expect_c[k];

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w0 = get_alpha_w(n);

        h_u_upd = hparam_u_upd[k][u];
        h_s_upd = hparam_s_upd[k][u];
        h_p_upd = hparam_p_upd[k][u];
        h_sdet_upd = hparam_sdet_upd[k][u];

        if (n > 1) { /* multivariate */

          sum = 0.0;
          for (j=1; j<=n; j++)
            sum += DIGAMMA(0.5 * (alpha_w0 + e_c_k + 1.0 - j));

          lp = 0.5 * (sum - LOG(h_sdet_upd) + n * LOG(2) - n / (tau + e_c_k));

          for (e=0; e<num_entry; e++) {
            n_miss = freq_list[e].num_miss_g[u];

            if (n_miss == 0) {  /* fully observed */
              x = alloc_vector(n);

              copy_vector(x,freq_list[e].gval[u]);
              diff_vector(x,h_u_upd,NULL);
              prob_comp[k][e] +=
                lp - 0.5 * (alpha_w0 + e_c_k) * qform(h_p_upd,x);

              release_vector(x);
            }
            else if (0 < n_miss && n_miss < n) { /* partially observed */
              n_obs = (n - n_miss);
              obs = freq_list[e].obs_g[u];

              x = alloc_vector(n_obs);
              h_u_upd_obs = alloc_vector(n_obs);
              h_s_upd_obs = alloc_matrix(n_obs,n_obs);
              h_s_upd_obs_inv = alloc_matrix(n_obs,n_obs);

              proj_vector(freq_list[e].gval[u],obs,x);
              proj_vector(h_u_upd,obs,h_u_upd_obs);
              proj_matrix(h_s_upd,obs,obs,h_s_upd_obs);
              h_sdet_upd_obs = INVERT(h_s_upd_obs,h_s_upd_obs_inv);
              if (err_inv)
                DIE_NUMERIC("Failed to invert the covariances of the"
                            " observed part\n");

              diff_vector(x,h_u_upd_obs,NULL);

              prob_comp[k][e] +=
                lp - 0.5 * (alpha_w0 + e_c_k) * qform(h_s_upd_obs_inv,x);
              prob_comp[k][e] -= 0.5 * (n - n_obs) * LOG(alpha_w0 + e_c_k);
              prob_comp[k][e] +=
                0.5 * (LOG(h_sdet_upd) - LOG(h_sdet_upd_obs) + (n - n_obs) * LOG(2 * M_PI));

              release_vector(x);
              release_vector(h_u_upd_obs);
              release_matrix(h_s_upd_obs);
              release_matrix(h_s_upd_obs_inv);
            }
            else { /* n_miss == n, i.e. all missing */
              prob_comp[k][e] +=
                lp - 0.5 * (n * LOG(alpha_w0 + e_c_k) - n * LOG(2 * M_PI) - LOG(h_sdet_upd));
            }
          }

        }
        else { /* univariate (n == 1) */
          lp = 0.5 * (DIGAMMA(0.5 * (alpha_w0 + e_c_k)) - LOG(h_sdet_upd) + LOG(2));
          lp -= 0.5 / (tau + e_c_k);

          for (e=0; e<num_entry; e++) {
            n_miss = freq_list[e].num_miss_g[u];

            if (n_miss == 0) {
              xval = freq_list[e].gval[u][0];
              prob_comp[k][e] += lp;
              prob_comp[k][e] -=
                0.5 * h_p_upd[0][0] * (alpha_w0 + e_c_k) *
                (xval - h_u_upd[0]) * (xval - h_u_upd[0]);
            }
            else {             /* all missing */
              prob_comp[k][e] += lp - 0.5 * (LOG(alpha_w0 + e_c_k) - LOG(2 * M_PI * h_sdet_upd));
            }
          }
        }
      }

    }
  }
}

void compute_conditional_prob_part(int *cand)
{
  int l,k,e;

  for (l=0; l<3; l++) {
    k = cand[l];
    for (e=0; e<num_entry; e++) {
      prob_cond[k][e] = EXP(prob_comp[k][e] - prob_incomp[e]);
      if (prob_cond[k][e] < PROB_COND_MIN) prob_cond[k][e] = PROB_COND_MIN;
    }
  }
}

void compute_expect_part(int *cand)
{
  int l,k,i,e,v,u;
  FREQ_PTR fp;
  FLOAT_T e_c_k,tmp_e;

  for (l=0; l<3; l++) {
    k = cand[l];
    e_c_k = 0.0;
#pragma omp parallel for reduction(+: e_c_k)
    for (e=0; e<num_entry; e++) {
      e_c_k += freq_list[e].freq * prob_cond[k][e];
    }
    expect_c[k] = e_c_k;
  }

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];

      for (v=0; v<tuple_size_d; v++)
        for (i=0; i<num_word[v]; i++)
          expect_w[k][v][i] = 0.0;

      /* Cannot immediately add a `pragma omp' statement here */
      for (e=0; e<num_entry; e++) {
        fp = &freq_list[e];
        tmp_e = fp->freq * prob_cond[k][e];
        for (v=0; v<tuple_size_d; v++)
          if (!fp->miss_w[v]) expect_w[k][v][fp->id[v]] += tmp_e;
      }
    }
  }

  if (mode_gauss) {
    for (l=0; l<3; l++) {
      k = cand[l];

      for (u=0; u<gauss_size; u++) {
        clear_vector(expect_u[k][u]);
        clear_matrix(expect_s[k][u]);
      }

      for (e=0; e<num_entry; e++) {
        tmp_e = freq_list[e].freq * prob_cond[k][e];
        compute_expect_gauss0_nodaem(k,e,tmp_e);
      }

      for (u=0; u<gauss_size; u++)
        div_const_vector(expect_u[k][u],expect_c[k]);
    }
  }

}

void update_hparam_part(int *cand)
{
  int k,l,i,v,u,n,a,a0,a1;
  FLOAT_T e_c_k,r,alpha_w0;
  VECTOR h_u_upd,e_u,h_u;
  MATRIX h_s_upd,e_s,h_s;
  FLOAT_T x_eu,x_es,x_hu,x_hs;

  for (l=0; l<3; l++) {
    k = cand[l];
    hparam_c_upd[k] = hparam_c[k] + expect_c[k];
  }

  if (mode_nominal) {
    for (l=0; l<3; l++) {
      k = cand[l];

      if (mode_nominal) {
#pragma omp parallel for private(i)
        for (v=0; v<tuple_size_d; v++) {
          for (i=0; i<num_word[v]; i++) {
            hparam_w_upd[k][v][i] = hparam_w[k][v][i] + expect_w[k][v][i];
          }
        }
      }
    }
  }

  if (mode_gauss) {
    for (l=0; l<3; l++) {
      k = cand[l];

      e_c_k = expect_c[k];
      r = (tau * e_c_k) / (tau + e_c_k);

      for (u=0; u<gauss_size; u++) {
        n = gauss_attr_size[u];
        alpha_w0 = get_alpha_w(n);

        h_u_upd = hparam_u_upd[k][u];
        h_s_upd = hparam_s_upd[k][u];

        if (n > 1) { /* multivariate */
          e_u = expect_u[k][u];
          e_s = expect_s[k][u];
          h_u = hparam_u[k][u];
          h_s = hparam_s[k][u];

          for (a=0; a<n; a++)
            h_u_upd[a] = (tau * h_u[a] + e_c_k * e_u[a])/(tau + e_c_k);

          for (a0=0; a0<n; a0++) {
            for (a1=a0; a1<n; a1++) {
              h_s_upd[a0][a1] = h_s[a0][a1];
              h_s_upd[a0][a1] += e_s[a0][a1] - e_c_k * e_u[a0] * e_u[a1];
              h_s_upd[a0][a1] += r * (h_u[a0] - e_u[a0]) * (h_u[a1] - e_u[a1]);
            }
          }
          copy_upper_to_lower(h_s_upd);
        }
        else { /* univariate */
          x_eu = expect_u[k][u][0];
          x_es = expect_s[k][u][0][0];
          x_hu = hparam_u[k][u][0];
          x_hs = hparam_s[k][u][0][0];

          h_u_upd[0] = (tau * x_hu + e_c_k * x_eu)/(tau + e_c_k);
          h_s_upd[0][0] =
            2 * x_hs + (x_es - e_c_k * x_eu * x_eu) + r * (x_hu - x_eu) * (x_hu - x_eu);
        }
      }
    }
  }

  if (mode_gauss) compute_hparam_precisions_part(cand);
}

void compute_incomplete_prob_part(int *cand)
{
  int l,k,e;
  FLOAT_T pcomp_max,sum;

  for (e=0; e<num_entry; e++) {
    pcomp_max = prob_incomp_part[e];
    for (l=0; l<3; l++) {
      k = cand[l];
      if (prob_comp[k][e] > pcomp_max) pcomp_max = prob_comp[k][e];
    }

    sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      sum += EXP(prob_comp[k][e] - pcomp_max);
    }
    sum += EXP(prob_incomp_part[e] - pcomp_max);

    prob_incomp[e] = pcomp_max + LOG(sum);
    /* print_msg(">> prob_incomp[%d] = %.Nf\n",e,prob_incomp[e]); */
  }
}

void save_free_energy_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    free_energy_best[m] = free_energy_sm[t][m];
}

void restore_free_energy_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    free_energy_sm[t][m] = free_energy_best[m];
}

void save_log_itemp_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp_best[m] = log_itemp_sm[t][m];
}

void restore_log_itemp_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++)
    log_itemp_sm[t][m] = log_itemp_best[m];
}

void save_hparam_sm(void)
{
  int k,v,i,u;

#pragma omp parallel for private(v,i,u)
  for (k=0; k<num_class; k++) {
    hparam_c_sm[k] = hparam_c[k];
    hparam_c_upd_sm[k] = hparam_c_upd[k];

    if (mode_nominal) {
      for (v=0; v<tuple_size_d; v++) {
        for (i=0; i<num_word[v]; i++) {
          hparam_w_sm[k][v][i] = hparam_w[k][v][i];
          hparam_w_upd_sm[k][v][i] = hparam_w_upd[k][v][i];
        }
      }
    }

    if (mode_gauss) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(hparam_u_sm[k][u],hparam_u[k][u]);
        copy_matrix(hparam_s_sm[k][u],hparam_s[k][u]);
        copy_matrix(hparam_p_sm[k][u],hparam_p[k][u]);
        hparam_sdet_sm[k][u] = hparam_sdet[k][u];

        copy_vector(hparam_u_upd_sm[k][u],hparam_u_upd[k][u]);
        copy_matrix(hparam_s_upd_sm[k][u],hparam_s_upd[k][u]);
        copy_matrix(hparam_p_upd_sm[k][u],hparam_p_upd[k][u]);
        hparam_sdet_upd_sm[k][u] = hparam_sdet_upd[k][u];
      }
      expect_c_sm[k] = expect_c[k];
    }
  }
}

void restore_hparam_sm(void)
{
  int k,v,i,u;

#pragma omp parallel for private(v,i,u)
  for (k=0; k<num_class; k++) {
    hparam_c[k] = hparam_c_sm[k];
    hparam_c_upd[k] = hparam_c_upd_sm[k];

    if (mode_nominal) {
      for (v=0; v<tuple_size_d; v++) {
        for (i=0; i<num_word[v]; i++) {
          hparam_w[k][v][i] = hparam_w_sm[k][v][i];
          hparam_w_upd[k][v][i] = hparam_w_upd_sm[k][v][i];
        }
      }
    }

    if (mode_gauss) {
      for (u=0; u<gauss_size; u++) {
        copy_vector(hparam_u[k][u],hparam_u_sm[k][u]);
        copy_matrix(hparam_s[k][u],hparam_s_sm[k][u]);
        copy_matrix(hparam_p[k][u],hparam_p_sm[k][u]);
        hparam_sdet[k][u] = hparam_sdet_sm[k][u];

        copy_vector(hparam_u_upd[k][u],hparam_u_upd_sm[k][u]);
        copy_matrix(hparam_s_upd[k][u],hparam_s_upd_sm[k][u]);
        copy_matrix(hparam_p_upd[k][u],hparam_p_upd_sm[k][u]);
        hparam_sdet_upd[k][u] = hparam_sdet_upd_sm[k][u];
      }
      expect_c[k] = expect_c_sm[k];
    }
  }
}

void save_incomplete_prob_part(int *cand)
{
  int k,e;
  int k1,k2,k3;
  FLOAT_T pcomp_max,sum;
  int first;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  for (e=0; e<num_entry; e++) {
    if (num_class <= 3) {
      prob_incomp_part[e] = 0.0;
      continue;
    }

    pcomp_max = 0.0;
    first = 1;
    for (k=0; k<num_class; k++) {
      if (k == k1 || k == k2 || k == k3) continue;
      if (first || prob_comp[k][e] > pcomp_max)	pcomp_max = prob_comp[k][e];
      first = 0;
    }

    sum = 0.0;
    for (k=0; k<num_class; k++) {
      if (k == k1 || k == k2 || k == k3) continue;
      sum += EXP(prob_comp[k][e] - pcomp_max);
    }

    prob_incomp_part[e] = pcomp_max + LOG(sum);
  }
}
