/*
 * NBCT
 * Copyright (c) 2007-2010, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef VNBC_DEF_H
#define VNBC_DEF_H

#ifndef VBNC
#define VBNC
#endif

#define FSX_DIST_VB    FSX_DIST
#define FSX_MEMP_VB    FSX_MEMP
#define FSX_LOG_VB     FSX_LOG
#define FSX_MSG_VB     FSX_MSG
#define FSX_CLUSTER_VB FSX_CLUSTER
#define FSX_RANK_VB    FSX_RANK
#define FSX_EVAL_VB    FSX_EVAL
#define FSX_PRED_VB    "predict"

#ifdef NO_LDOUBLE
#  define DEFAULT_TAU (1.0e-3)
#  define PROB_COND_MIN (1.0e-30)
#else
#  define DEFAULT_TAU (1.0e-3L)
#  define PROB_COND_MIN (1.0e-30L)
#endif

enum vnbc_option_id {
  O_VB_N_CAND=1, O_VB_DAEM, O_VB_ITEMP, O_VB_SMOOTH_C,
  O_VB_MIN_NEW_C_PARAM, O_VB_DISSIM, O_VB_MEM_PROB, O_VB_EPSILON, O_VB_EVAL,
  O_VB_TARGET, O_VB_NORM_JMERGE, O_VB_N_JSPLIT, O_VB_HELP, O_VB_INPUT,
  O_VB_OUTPUT, O_VB_N_CLASS, O_VB_SMOOTH_MEAN, O_VB_MAX_ITER_RESTART,
  O_VB_SMOOTH_VAR, O_VB_MAX_ITER, O_VB_SAVE_MEMORY, O_VB_RESTART,
  O_VB_MIN_SM_IMP, O_VB_N_SM, O_VB_ALPHA, O_VB_CODE_IN, O_VB_CODE_OUT,
  O_VB_SEED, O_VB_READ_HPARAM, O_VB_SUFFIX, O_VB_SAVE_MESSAGE, O_VB_TAU,
  O_VB_GAUSS, O_VB_MIN_C_PARAM, O_VB_VERBOSE, O_VB_SMOOTH_A,
  O_VB_N_ATTR, O_VB_OCC_ORDER, O_VB_RANK, O_VB_ID, O_VB_IGNORE,
  O_VB_XHELP, O_VB_HIDE_ATTR, O_VB_RANK_HIDE, O_VB_PREC_OUT,
  O_VB_PREC_MSG, O_VB_UNCOMPRESS, O_VB_W_PURITY, O_VB_W_WALLANCE,
  O_VB_SHOW_ATTR_MEMP, O_VB_SHOW_ATTR_CLUSTER, O_VB_VERSION, O_VB_REGULAR_STAT,
  O_VB_NOISE_RATIO, O_VB_NOISE_RATIO_GAUSS, O_VB_EXTREME, O_VB_PREDICT,
  O_VB_NOOP
};

#define OPTSTR_VB(i)  get_option_str((i),vnbc_opts,0)
#define LOPTSTR_VB(i) get_option_str((i),vnbc_opts,1)

/* vbem.c */
void vbem_main(void);
void no_vbem_main(void);
void vbem_iterate(int);
int vbem_iterate_body(int,int,int,FLOAT_T);
FLOAT_T compute_free_energy(void);
FLOAT_T compute_free_energy_daem(void);
FLOAT_T compute_predict_gauss(int,int,int);
void update_hparam(void);
void update_param_vbem(void);
void update_param_no_vbem(void);
void initialize_hparam(void);
void compute_pi(void);
void save_free_energy_best(int);
void restore_free_energy_best(int);
void save_hparam_opt(void);
void restore_hparam_opt(void);
void save_hparam_best(void);
void restore_hparam_best(void);
/* vbsmem.c */
void vbsmem_iterate(int);
int vbsmem_iterate_full(int,int,int,FLOAT_T);
int vbsmem_iterate_part(int,int,int,int);
void compute_jsplit(void);
void compute_jmerge(void);
void initialize_hparam_part(int *);
void update_hparam_part(int *);
void compute_pi_part(int *);
void compute_complete_prob_part(int *);
void compute_incomplete_prob_part(int *);
void compute_conditional_prob_part(int *);
void compute_expect_prob_part(int *);
void sort_candidate(void);
void save_hparam_sm(void);
void save_free_energy_best_sm(int,int);
void save_log_itemp_best_sm(int,int);
void restore_hparam_sm(void);
void restore_free_energy_best_sm(int,int);
void restore_log_itemp_best_sm(int,int);
/* read_param_vb.c */
void read_hparam(void);
void copy_read_hparam(void);
/* allocate_vb.c */
void alloc_seed_vb(void);
void alloc_hparam_upd(void);
void alloc_hparam_best(void);
void alloc_hparam_sm(void);
void alloc_hparam_vb(void);
void alloc_free_energy(void);
void alloc_free_energy_best(void);
void alloc_free_energy_sm(void);
void alloc_free_energy_sm_size(void);
void alloc_log_itemp_vb(void);
void alloc_log_itemp_sm_vb(void);
void alloc_pi(void);
void expand_free_energy_all(void);
void expand_free_energy_sm_all(int);
void release_seed_vb(void);
void release_hparam_upd(void);
void release_hparam_best(void);
void release_hparam_opt(void);
void release_hparam_sm(void);
void release_hparam_vb(void);
void release_free_energy(void);
void release_free_energy_best(void);
void release_free_energy_sm(void);
void release_free_energy_sm_size(void);
void release_log_itemp_vb(void);
void release_log_itemp_sm_vb(void);
void release_pi(void);
/* cluster_vb.c */
void cluster_main_vb(void);
/* rank_vb.c */
void rank_main_vb(void);
/* print_vb.c */
void print_time_vb(FILE *);
void print_hparam(FILE *);
void print_hparam_upd(FILE *);
void print_pi(FILE *);
void print_free_energy(FILE *,int,int);
void print_free_energy_sm(FILE *,int);
void print_ic(FILE *);
void print_predict_prob(FILE *);

#endif
